"""FAST Pinball mock hardware platform. In this module:

MockFastSerial
MockFastNetNeuron
MockFastExp
MockFastRgb
MockFastNetNano
MockFastSeg
MockFastRgbDmd
"""

import re
from mpf.tests.loop import MockSerial

class MockFastSerial(MockSerial):

    PRINT_FSP_TRAFFIC = False

    def __init__(self, test_fast_base):
        super().__init__()
        self.test_fast_base = test_fast_base
        self.type = None
        self.queue = []
        self.msg_history = list()  # list of commands received, e.g. ['ID:', 'ID@88:', 'ID@89:', 'EA:880', 'RD:0200ffffff121212']
        self.expected_commands = dict()  # popped when called, verify empty at end of test
        self.autorespond_commands = dict()  # can be called multiple times, never popped
        self.port = None

    def read(self, length):
        """ Reads the message from receive queue"""
        del length
        if not self.queue:
            return

        msg = self.queue.pop()
        msg = (msg + '\r').encode()
        return msg

    def read_ready(self):
        return bool(len(self.queue) > 0)

    def write_ready(self):
        return True

    def process_msg(self, msg):
        """Override this to apply further processing to a message.
        Return True if processed and complete, False if not."""
        return False

    def write(self, msg):
        """Write message out to the serial port."""
        lines = msg.split(b'\r')

        for line in lines:
            if line == b'':
                continue

            # Split based on 'RD@xxx:' pattern (hex included)
            parts = re.split(b'(RD@[0-9A-Fa-f]+:)', line)

            # re adds an empty string to the beginning of the list if the string starts with the pattern
            if parts and parts[0] == b'':
                parts.pop(0)

            # Reassemble the parts, taking binary into consideration
            for i in range(0, len(parts) - 1, 2):
                prefix = parts[i]  # RD@xxx:
                payload = parts[i + 1]  # binary payload

                # Convert binary payload back to hex chars
                hex_payload = payload.hex()

                # Assemble back the message part with the prefix
                complete_message = f"{prefix.decode('utf-8', 'ignore')}{hex_payload}"

                if complete_message == '':
                    continue

                # Simulate board response with the reassembled message
                self._simulate_board_response(complete_message.encode('utf-8'))

            # If the line did not contain 'RD@xxx:', treat it as a regular ASCII message
            if len(parts) == 1:
                self._simulate_board_response(line)

        return len(msg)

    def _simulate_board_response(self, msg):
        # handles the processing of outgoing messages

        msg_len = len(msg)
        try:
            cmd = msg.decode()
        except UnicodeDecodeError:
            cmd = self._process_binary_msg(msg)

        if self.PRINT_FSP_TRAFFIC:
            print(f'{self.type} >>> {cmd}')

        self.msg_history.append(cmd)

        # Order of precedence is key here. We want autoresponses to be last so they can be overridden
        # via custom responses or expected commands on a per-test basis.

        rsp = self.process_msg(cmd)

        if type(rsp) is str and rsp:
            if self.PRINT_FSP_TRAFFIC:
                print(f'{self.type} <<< {rsp}')
            self.queue.append(rsp)
            return msg_len

        if rsp == True:
            return 0

        if cmd in self.expected_commands:
            if self.expected_commands[cmd]:
                self.queue.append(self.expected_commands[cmd])
                if self.PRINT_FSP_TRAFFIC:
                    print(f'{self.type} <<< {self.expected_commands[cmd]}')
            del self.expected_commands[cmd]
            return msg_len

        if cmd in self.autorespond_commands:
            self.queue.append(self.autorespond_commands[cmd])
            if self.PRINT_FSP_TRAFFIC:
                print(f'{self.type} <<< {self.autorespond_commands[cmd]}')
            return msg_len

        raise Exception(f"Unexpected command for {self.type}: {cmd}")

    def _process_binary_msg(self, msg):
        # The first three chars are the command (XX:), the rest is the binary payload
        return f'{msg[:3].decode()}{msg[3:].hex()}'


class MockFastNetNeuron(MockFastSerial):
    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = "NETv2"
        self.port = 'com3'

        self.autorespond_commands = {
            'WD:1' : 'WD:P',
            'WD:3E8': 'WD:P',
            'SA:':'SA:0E,2900000000000000000000000000',
            'CH:2000,FF':'CH:P',
            'ID:': 'ID:NET FP-CPU-2000  02.13',
            'BR:': '\r\r!B:00\r..!B:02\r.',
            }

        self.attached_boards = dict()
        self.msg_history = list()

    def process_msg(self, cmd):
        if cmd == (' ' * 256 * 4):
            return "XX:F"


class MockFastExp(MockFastSerial):
    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = 'EXP'
        self.port = 'com4'
        self.leds = dict()
        self.led_map = dict()  # LED number to name index, e.g. 88000: "led1", 88121: "led5"

    def process_msg(self, cmd):
        # returns True if the msg was fully processed, False if it was not

        try:
            cmd, payload = cmd.split(":", 1)
        except ValueError:  # binary encoded message
            cmd, payload = bytes.fromhex(cmd).decode().split(":", 1)

        cmd, address = cmd.split("@", 1)

        if cmd == "ID":
            if not address:  # no ID has been set, so lowest address will respond
                return "ID:EXP FP-EXP-2000 0.11"

            if address == '48':  # Neuron
                return "ID:EXP FP-EXP-2000 0.11"

            elif address in ["B4", "B5", "B6", "B7"]:  # 71
                return "ID:EXP FP-EXP-0071  0.11"

            elif address in ["84", "85", "86", "87"]:  # 81
                return "ID:EXP FP-EXP-0081  0.12"

            elif address in ["88", "89", "8A", "8B"]:  # 91
                return "ID:EXP FP-EXP-0091  0.11"

            # Breakouts
            elif address == "480":  # Neuron
                return "ID:LED FP-BRK-0001  0.8"

            elif address == '481':  # Neuron
                return "ID:BRK FP-PWR-0007  0.8"

            elif address == '482':  # Neuron
                return "ID:BRK FP-BRK-0116  0.8"

            elif address in ["B40", "B50", "B60", "B70"]:  # 71
                return "ID:BRK FP-EXP-0071  0.11"

            elif address in ["840", "850", "860", "870",
                                 "841", "851", "861", "871"]:  # 81
                return "ID:BRK FP-EXP-0081  0.12"

            elif address in ["880", "890", "8A0", "8B0"]:  # 91
                return "ID:BRK FP-EXP-0091  0.11"

            elif address in ["881", "892"]:  # 091
                return "ID:BRK FP-BRK-0001  0.8"

            elif address in ["882"]:  # 091
                return "ID:BRK FP-DRV-0800  0.0"

            assert False, f"Unexpected ID request for {address}"

        elif cmd == "BR":
            # turn off all the LEDs on that board
            for led_number, led_name in self.led_map.items():
                if led_number.startswith(address):
                    self.leds[led_name] = "000000"
            return "BR:P"

        elif cmd == "RD":
            # LED color, update our map of LED colors
            if not self.led_map:
                self._create_led_map()

            payload = payload.upper()
            count = int(payload[:2], 16)
            color_data = payload[2:]

            assert len(color_data) == count * 8

            # update our record of the LED colors
            for i in range(count):
                color = color_data[i * 8 + 2:i * 8 + 8]
                led_number = f'{address}{color_data[i * 8:i * 8 + 2]}'

                self.leds[self.led_map[led_number]] = color

        elif cmd == "RA":
            # LED color, update all LEDs
            if not self.led_map:
                return  # no map means nothing to reset yet

            for led_number, led_name in self.led_map.items():
                if led_number.startswith(address):
                    self.leds[led_name] = payload.upper()

    def _create_led_map(self):
        # This just gives us a map of LED numbers to LED names, e.g. 88000: "led1", 88121: "led5"
        # so we can reference them by name in our tests. It's a bit awkward when logical LEDs
        # span multiple physical LEDs, but meh, it's just for tests

        for name, led in self.test_fast_base.machine.lights.items():
            led_number = led.hw_drivers['red'][0].number.split('-')[0]
            self.led_map[led_number] = name
            led_number = led.hw_drivers['green'][0].number.split('-')[0]
            self.led_map[led_number] = name
            led_number = led.hw_drivers['blue'][0].number.split('-')[0]
            self.led_map[led_number] = name
            try:
                led_number = led.hw_drivers['white'][0].number.split('-')[0]
                self.led_map[led_number] = name
            except KeyError:
                pass

class MockFastRgb(MockFastSerial):
    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = 'RGB'
        self.port = 'com5'

        self.autorespond_commands = {
            'ID:': 'ID:RGB FP-CPU-002-2 01.00',
            }

        self.leds = dict()
        for i in range(128):
            self.leds[f'{i:02X}'] = "000000"

    def process_msg(self, cmd):
        if cmd[:3] == 'RS:':
            remaining = cmd[3:]
            while True:
                self.leds[remaining[0:2]] = remaining[2:8]
                remaining = remaining[9:]

                if not remaining:
                    break

            return "RX:P"

        elif cmd[:3] == 'RA:':
            for led in self.leds:
                self.leds[led] = cmd[3:9]

            return "RX:P"

class MockFastNetNano(MockFastSerial):
    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = "NETv1"
        self.port = 'com6'

        self.autorespond_commands = {
            'WD:1' : 'WD:P',
            'WD:3E8': 'WD:P',
            "SA:": "SA:01,00,09,000000040000000000",  # switch 0x1A is active
            'ID:': 'ID:NET FP-CPU-002-2  01.05',
            }

        self.attached_boards = dict()
        self.msg_history = list()

    def process_msg(self, cmd):
        if cmd == (' ' * 256 * 4):
            return "XX:F"

class MockFastNetRetro(MockFastSerial):
    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = "NETv2"
        self.port = 'com3'

        self.autorespond_commands = {
            'WD:1' : 'WD:P',
            'WD:3E8': 'WD:P',
            'SA:':'SA:0E,2900000000000000000000000000',
            'CH:9500,FF':'CH:P',
            'ID:': 'ID:NET FP-SBI-0095  02.13',
            'BR:': '\r\r!B:00\r..!B:02\r.',
            }

        self.attached_boards = dict()
        self.msg_history = list()

    def process_msg(self, cmd):
        if cmd == (' ' * 256 * 4):
            return "XX:F"

class MockFastSeg(MockFastSerial):

    PRINT_FSP_TRAFFIC = False

    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = "SEG"
        self.port = "com7"
        self.seg_digits = dict()
        self.seg_colors = dict()

        self.autorespond_commands = {
            'ID:': 'ID:SEG FP-CPU-002-2 00.10',
            }

    def process_msg(self, cmd):
        cmd = cmd.upper()
        cmd, payload = cmd.split(':')

        if cmd == 'PA':
            idx, data = payload.split(',')
            self.seg_digits[idx] = data
            return True  # processed, no return msg

        # PC:00,ffffff,ffffff,ffffff,ffffff,ffffff,ffffff,ffffff,
        elif cmd == 'PC':
            sliced = payload.split(',')
            idx = sliced.pop(0)
            # A training comma will lead to an extra list item we don't want
            if not sliced[-1:][0]:
                sliced.pop(-1)

            self.seg_colors[idx] = sliced
            return True

class MockFastRgbDmd(MockFastSerial):

    PRINT_FSP_TRAFFIC = False

    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = "DMD"
        self.port = 'com8'

        self.autorespond_commands = {
            'ID:': 'ID:DMD FP-CPU-002-2 00.88',
            }

    def process_msg(self, cmd):
        md = cmd.upper()
        cmd, payload = cmd.split(':')

        if cmd == 'BM':
            assert len(payload) == 128 * 32 * 3
            return 'BM:P'

class MockFastAudio(MockFastSerial):

    PRINT_FSP_TRAFFIC = False

    def __init__(self, test_fast_base):
        super().__init__(test_fast_base)
        self.type = "AUD"
        self.port = "com9"
        self.main_volume = '0'
        self.sub_volume = '0'
        self.headphones_volume = '0'
        self.board_config_byte = '00'

        self.autorespond_commands = {
            'ID:': 'ID:AUD FP-AUD-0100 00.10',
            }

    def process_msg(self, cmd):
        cmd = cmd.upper()
        if cmd[:2] in ('AV', 'AS', 'AH'):
            return self.process_volume_command(cmd)
        elif cmd[:2] == 'AM':
            return self.process_am_command(cmd)

    def process_volume_command(self, cmd):
        device, volume = cmd.split(':')

        if len(volume) == 1:
            volume = f'0{volume}'
        assert len(volume) == 2
        assert volume >= '00'
        assert volume <= '3F'

        if device == 'AV':
            self.main_volume = volume
        elif device == 'AS':
            self.sub_volume = volume
        elif device == 'AH':
            self.headphones_volume = volume
        else:
            raise AssertionError(f'Unknown device {device}')

    def process_am_command(self, cmd):
        self.board_config_byte = cmd[3:]


# mpf.tests.test_Fast_Nano

from mpf.core.rgb_color import RGBColor

from mpf.tests.MpfTestCase import MagicMock, test_config, expect_startup_error
from mpf.tests.test_Fast import TestFastBase


class TestFastNano(TestFastBase):
    """FAST Platform class for a networked V1 platform. Tests the NET v1 and RGB processors."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serial_connections_to_mock = ['net1', 'rgb']

    def get_config_file(self):
        return 'nano.yaml'

    def create_expected_commands(self):

        self.net_cpu.expected_commands = {
            "SN:00": "SN:00,01,1E,1E",
            "SN:01": "SN:01,01,1E,1E",
            "SN:02": "SN:02,01,1E,1E",
            "SN:03": "SN:03,01,1E,1E",
            "SN:04": "SN:04,01,1E,1E",
            "SN:05": "SN:05,01,1E,1E",
            "SN:06": "SN:06,01,1E,1E",
            "SN:07": "SN:07,01,1E,1E",
            "SN:08": "SN:08,01,1E,1E",
            "SN:09": "SN:09,01,1E,1E",
            "SN:0A": "SN:0A,01,1E,1E",
            "SN:0B": "SN:0B,01,1E,1E",
            "SN:0C": "SN:0C,01,1E,1E",
            "SN:0D": "SN:0D,01,1E,1E",
            "SN:0E": "SN:0E,01,1E,1E",
            "SN:0F": "SN:0F,01,1E,1E",
            "SN:10": "SN:10,01,1E,1E",
            "SN:11": "SN:11,01,1E,1E",
            "SN:12": "SN:12,01,1E,1E",
            "SN:13": "SN:13,01,1E,1E",
            "SN:14": "SN:14,01,1E,1E",
            "SN:15": "SN:15,01,1E,1E",
            "SN:16": "SN:16,01,1E,1E",
            "SN:17": "SN:17,01,1E,1E",
            "SN:18": "SN:18,01,1E,1E",
            "SN:19": "SN:19,01,1E,1E",
            "SN:1A": "SN:1A,01,1E,1E",
            "SN:1B": "SN:1B,01,1E,1E",
            "SN:1C": "SN:1C,01,1E,1E",
            "SN:1D": "SN:1D,01,1E,1E",
            "SN:1E": "SN:1E,01,1E,1E",
            "SN:1F": "SN:1F,01,1E,1E",
            "SN:20": "SN:20,01,1E,1E",
            "SN:21": "SN:21,01,1E,1E",
            "SN:22": "SN:22,01,1E,1E",
            "SN:23": "SN:23,01,1E,1E",
            "SN:24": "SN:24,01,1E,1E",
            "SN:25": "SN:25,01,1E,1E",
            "SN:26": "SN:26,01,1E,1E",
            "SN:27": "SN:27,01,1E,1E",
            "SN:28": "SN:28,01,1E,1E",
            "SN:29": "SN:29,01,1E,1E",
            "SN:2A": "SN:2A,01,1E,1E",
            "SN:2B": "SN:2B,01,1E,1E",
            "SN:2C": "SN:2C,01,1E,1E",
            "SN:2D": "SN:2D,01,1E,1E",
            "SN:2E": "SN:2E,01,1E,1E",
            "SN:2F": "SN:2F,01,1E,1E",
            "SN:30": "SN:30,01,1E,1E",
            "SN:31": "SN:31,01,1E,1E",
            "SN:32": "SN:32,01,1E,1E",
            "SN:33": "SN:33,01,1E,1E",
            "SN:34": "SN:34,01,1E,1E",
            "SN:35": "SN:35,01,1E,1E",
            "SN:36": "SN:36,01,1E,1E",
            "SN:37": "SN:37,01,1E,1E",
            "SN:38": "SN:38,01,1E,1E",
            "SN:39": "SN:39,01,1E,1E",
            "SN:3A": "SN:3A,01,1E,1E",
            "SN:3B": "SN:3B,01,1E,1E",
            "SN:3C": "SN:3C,01,1E,1E",
            "SN:3D": "SN:3D,01,1E,1E",
            "SN:3E": "SN:3E,01,1E,1E",
            "SN:3F": "SN:3F,01,1E,1E",
            "SN:40": "SN:40,01,1E,1E",
            "SN:41": "SN:41,01,1E,1E",
            "SN:42": "SN:42,01,1E,1E",
            "SN:43": "SN:43,01,1E,1E",
            "SN:44": "SN:44,01,1E,1E",
            "SN:45": "SN:45,01,1E,1E",
            "SN:46": "SN:46,01,1E,1E",
            "SN:47": "SN:47,01,1E,1E",
            "SN:48": "SN:48,01,1E,1E",
            "SN:49": "SN:49,01,1E,1E",
            "SN:4A": "SN:4A,01,1E,1E",
            "SN:4B": "SN:4B,01,1E,1E",
            "SN:4C": "SN:4C,01,1E,1E",
            "SN:4D": "SN:4D,01,1E,1E",
            "SN:4E": "SN:4E,01,1E,1E",
            "SN:4F": "SN:4F,01,1E,1E",
            "SN:50": "SN:50,01,1E,1E",
            "SN:51": "SN:51,01,1E,1E",
            "SN:52": "SN:52,01,1E,1E",
            "SN:53": "SN:53,01,1E,1E",
            "SN:54": "SN:54,01,1E,1E",
            "SN:55": "SN:55,01,1E,1E",
            "SN:56": "SN:56,01,1E,1E",
            "SN:57": "SN:57,01,1E,1E",
            "SN:58": "SN:58,01,1E,1E",
            "SN:59": "SN:59,01,1E,1E",
            "SN:5A": "SN:5A,01,1E,1E",
            "SN:5B": "SN:5B,01,1E,1E",
            "SN:5C": "SN:5C,01,1E,1E",
            "SN:5D": "SN:5D,01,1E,1E",
            "SN:5E": "SN:5E,01,1E,1E",
            "SN:5F": "SN:5F,01,1E,1E",
            "SN:60": "SN:60,01,1E,1E",
            "SN:61": "SN:61,01,1E,1E",
            "SN:62": "SN:62,01,1E,1E",
            "SN:63": "SN:63,01,1E,1E",
            "SN:64": "SN:64,01,1E,1E",
            "SN:65": "SN:65,01,1E,1E",
            "SN:66": "SN:66,01,1E,1E",
            "SN:67": "SN:67,01,1E,1E",
            "SN:68": "SN:68,01,1E,1E",
            "SN:69": "SN:69,01,1E,1E",
            "SN:6A": "SN:6A,01,1E,1E",
            "SN:6B": "SN:6B,01,1E,1E",

            # Initial driver responses before they're configured:
            "DN:00": "DN:00,00,00,00,00,00,00,00,00",
            "DN:01": "DN:01,00,00,00,00,00,00,00,00",
            "DN:02": "DN:02,00,00,00,00,00,00,00,00",
            "DN:03": "DN:03,00,00,00,00,00,00,00,00",
            "DN:04": "DN:04,00,00,00,00,00,00,00,00",
            "DN:05": "DN:05,00,00,00,00,00,00,00,00",
            "DN:06": "DN:06,00,00,00,00,00,00,00,00",
            "DN:07": "DN:07,00,00,00,00,00,00,00,00",
            "DN:08": "DN:08,00,00,00,00,00,00,00,00",
            "DN:09": "DN:09,00,00,00,00,00,00,00,00",
            "DN:0A": "DN:0A,00,00,00,00,00,00,00,00",
            "DN:0B": "DN:0B,00,00,00,00,00,00,00,00",
            "DN:0C": "DN:0C,00,00,00,00,00,00,00,00",
            "DN:0D": "DN:0D,00,00,00,00,00,00,00,00",
            "DN:0E": "DN:0E,00,00,00,00,00,00,00,00",
            "DN:0F": "DN:0F,00,00,00,00,00,00,00,00",
            "DN:10": "DN:10,00,00,00,00,00,00,00,00",
            "DN:11": "DN:11,00,00,00,00,00,00,00,00",
            "DN:12": "DN:12,00,00,00,00,00,00,00,00",
            "DN:13": "DN:13,00,00,00,00,00,00,00,00",
            "DN:14": "DN:14,00,00,00,00,00,00,00,00",
            "DN:15": "DN:15,00,00,00,00,00,00,00,00",
            "DN:16": "DN:16,00,00,00,00,00,00,00,00",
            "DN:17": "DN:17,00,00,00,00,00,00,00,00",
            "DN:18": "DN:18,00,00,00,00,00,00,00,00",
            "DN:19": "DN:19,00,00,00,00,00,00,00,00",
            "DN:1A": "DN:1A,00,00,00,00,00,00,00,00",
            "DN:1B": "DN:1B,00,00,00,00,00,00,00,00",
            "DN:1C": "DN:1C,00,00,00,00,00,00,00,00",
            "DN:1D": "DN:1D,00,00,00,00,00,00,00,00",
            "DN:1E": "DN:1E,00,00,00,00,00,00,00,00",
            "DN:1F": "DN:1F,00,00,00,00,00,00,00,00",
            "DN:20": "DN:20,00,00,00,00,00,00,00,00",
            "DN:21": "DN:21,00,00,00,00,00,00,00,00",
            "DN:22": "DN:22,00,00,00,00,00,00,00,00",
            "DN:23": "DN:23,00,00,00,00,00,00,00,00",
            "DN:24": "DN:24,00,00,00,00,00,00,00,00",
            "DN:25": "DN:25,00,00,00,00,00,00,00,00",
            "DN:26": "DN:26,00,00,00,00,00,00,00,00",
            "DN:27": "DN:27,00,00,00,00,00,00,00,00",
            "DN:28": "DN:28,00,00,00,00,00,00,00,00",
            "DN:29": "DN:29,00,00,00,00,00,00,00,00",
            "DN:2A": "DN:2A,00,00,00,00,00,00,00,00",
            "DN:2B": "DN:2B,00,00,00,00,00,00,00,00",
            "DN:2C": "DN:2C,00,00,00,00,00,00,00,00",
            "DN:2D": "DN:2D,00,00,00,00,00,00,00,00",
            "DN:2E": "DN:2E,00,00,00,00,00,00,00,00",
            "DN:2F": "DN:2F,00,00,00,00,00,00,00,00",

            'NN:00': 'NN:00,FP-I/O-3208-2   ,01.05,08,20,04,06,00,00,00,00',
            'NN:01': 'NN:01,FP-I/O-0804-1   ,01.05,04,08,04,06,00,00,00,00',
            'NN:02': 'NN:02,FP-I/O-1616-2   ,01.05,10,10,04,06,00,00,00,00',
            'NN:03': 'NN:03,FP-I/O-1616-2   ,01.05,10,10,04,06,00,00,00,00',

            # Initialization commands, also tests all the various switch config options
            "SN:00,00,00,00": "SN:P",
            "SN:01,01,04,04": "SN:P",
            "SN:02,01,04,04": "SN:P",
            "SN:03,01,04,04": "SN:P",
            "SN:04,00,00,00": "SN:P",
            "SN:05,00,00,00": "SN:P",
            "SN:06,00,00,00": "SN:P",
            "SN:07,01,05,1A": "SN:P",
            "SN:08,00,00,00": "SN:P",
            "SN:09,00,00,00": "SN:P",
            "SN:0A,00,00,00": "SN:P",
            "SN:0B,01,04,04": "SN:P",
            "SN:0C,01,04,04": "SN:P",
            "SN:0D,00,00,00": "SN:P",
            "SN:0E,00,00,00": "SN:P",
            "SN:0F,00,00,00": "SN:P",
            "SN:10,00,00,00": "SN:P",
            "SN:11,00,00,00": "SN:P",
            "SN:12,00,00,00": "SN:P",
            "SN:13,00,00,00": "SN:P",
            "SN:14,00,00,00": "SN:P",
            "SN:15,00,00,00": "SN:P",
            "SN:16,01,04,04": "SN:P",
            "SN:17,00,00,00": "SN:P",
            "SN:18,00,00,00": "SN:P",
            "SN:19,00,00,00": "SN:P",
            "SN:1A,02,04,04": "SN:P",
            "SN:1B,00,00,00": "SN:P",
            "SN:1C,00,00,00": "SN:P",
            "SN:1D,00,00,00": "SN:P",
            "SN:1E,00,00,00": "SN:P",
            "SN:1F,00,00,00": "SN:P",
            "SN:20,00,00,00": "SN:P",
            "SN:21,00,00,00": "SN:P",
            "SN:22,00,00,00": "SN:P",
            "SN:23,00,00,00": "SN:P",
            "SN:24,00,00,00": "SN:P",
            "SN:25,00,00,00": "SN:P",
            "SN:26,00,00,00": "SN:P",
            "SN:27,00,00,00": "SN:P",
            "SN:28,00,00,00": "SN:P",
            "SN:29,00,00,00": "SN:P",
            "SN:2A,00,00,00": "SN:P",
            "SN:2B,00,00,00": "SN:P",
            "SN:2C,00,00,00": "SN:P",
            "SN:2D,00,00,00": "SN:P",
            "SN:2E,00,00,00": "SN:P",
            "SN:2F,00,00,00": "SN:P",
            "SN:30,00,00,00": "SN:P",
            "SN:31,00,00,00": "SN:P",
            "SN:32,00,00,00": "SN:P",
            "SN:33,00,00,00": "SN:P",
            "SN:34,00,00,00": "SN:P",
            "SN:35,00,00,00": "SN:P",
            "SN:36,00,00,00": "SN:P",
            "SN:37,00,00,00": "SN:P",
            "SN:38,00,00,00": "SN:P",
            "SN:39,01,04,04": "SN:P",
            "SN:3A,00,00,00": "SN:P",
            "SN:3B,00,00,00": "SN:P",
            "SN:3C,00,00,00": "SN:P",
            "SN:3D,00,00,00": "SN:P",
            "SN:3E,00,00,00": "SN:P",
            "SN:3F,00,00,00": "SN:P",
            "SN:40,00,00,00": "SN:P",
            "SN:41,00,00,00": "SN:P",
            "SN:42,00,00,00": "SN:P",
            "SN:43,00,00,00": "SN:P",
            "SN:44,00,00,00": "SN:P",
            "SN:45,00,00,00": "SN:P",
            "SN:46,00,00,00": "SN:P",
            "SN:47,00,00,00": "SN:P",
            "SN:48,00,00,00": "SN:P",
            "SN:49,00,00,00": "SN:P",
            "SN:4A,00,00,00": "SN:P",
            "SN:4B,00,00,00": "SN:P",
            "SN:4C,00,00,00": "SN:P",
            "SN:4D,00,00,00": "SN:P",
            "SN:4E,00,00,00": "SN:P",
            "SN:4F,00,00,00": "SN:P",
            "SN:50,00,00,00": "SN:P",
            "SN:51,00,00,00": "SN:P",
            "SN:52,00,00,00": "SN:P",
            "SN:53,00,00,00": "SN:P",
            "SN:54,00,00,00": "SN:P",
            "SN:55,00,00,00": "SN:P",
            "SN:56,00,00,00": "SN:P",
            "SN:57,00,00,00": "SN:P",
            "SN:58,00,00,00": "SN:P",
            "SN:59,00,00,00": "SN:P",
            "SN:5A,00,00,00": "SN:P",
            "SN:5B,00,00,00": "SN:P",
            "SN:5C,00,00,00": "SN:P",
            "SN:5D,00,00,00": "SN:P",
            "SN:5E,00,00,00": "SN:P",
            "SN:5F,00,00,00": "SN:P",
            "SN:60,00,00,00": "SN:P",
            "SN:61,00,00,00": "SN:P",
            "SN:62,00,00,00": "SN:P",
            "SN:63,00,00,00": "SN:P",
            "SN:64,00,00,00": "SN:P",
            "SN:65,00,00,00": "SN:P",
            "SN:66,00,00,00": "SN:P",
            "SN:67,00,00,00": "SN:P",
            "SN:68,00,00,00": "SN:P",
            "SN:69,00,00,00": "SN:P",
            "SN:6A,00,00,00": "SN:P",
            "SN:6B,00,00,00": "SN:P",

            # Drivers with configs, tests config options
            "DN:01,81,00,10,FF,FF,00,FF,00": "DN:P",  # initial digital output config, will be updated later
            "DN:04,81,00,10,17,FF,00,00,1B": "DN:P",
            "DN:06,81,00,10,17,FF,00,FF,00": "DN:P",
            "DN:07,81,00,10,0A,FF,00,00,00": "DN:P",
            "DN:11,81,00,10,0A,AA,00,92,00": "DN:P",
            "DN:12,81,00,70,0A,FF,14,FF,00": "DN:P",
            "DN:13,81,00,10,0A,FF,00,88,00": "DN:P",
            "DN:16,81,00,10,14,FF,00,88,00": "DN:P",
            "DN:17,81,00,10,14,FF,00,88,00": "DN:P",
            "DN:20,81,00,10,0A,FF,00,01,00": "DN:P",
            "DN:21,81,00,10,0A,FF,00,01,00": "DN:P",

            # Digital output which after the initial pass
            "DN:01,C1,00,18,00,FF,FF,00,00": "DN:P",
            }

    def setUp(self):
        super().setUp()
        if not self.startup_error:
            self.advance_time_and_run()
            self.assertFalse(self.net_cpu.expected_commands)
            self.assertFalse(self.rgb_cpu.expected_commands)

            # test io board detection
            self.assertEqual(4, len(self.machine.default_platform.io_boards))
            self.assertEqual(32, self.machine.default_platform.io_boards[0].switch_count)
            self.assertEqual(8, self.machine.default_platform.io_boards[0].driver_count)
            self.assertEqual(8, self.machine.default_platform.io_boards[1].switch_count)
            self.assertEqual(4, self.machine.default_platform.io_boards[1].driver_count)
            self.assertEqual(16, self.machine.default_platform.io_boards[2].switch_count)
            self.assertEqual(16, self.machine.default_platform.io_boards[2].driver_count)
            self.assertEqual(16, self.machine.default_platform.io_boards[3].switch_count)
            self.assertEqual(16, self.machine.default_platform.io_boards[3].driver_count)

    def test_coils(self):
        self._test_pulse()
        self._test_long_pulse()
        self._test_timed_enable()
        self._test_default_timed_enable()
        self._test_enable_exception()
        self._test_allow_enable()
        self._test_pwm_ssm()

        # test hardware scan
        info_str = (
            'NET: FP-CPU-002-2 v01.05\n'
            'RGB: FP-CPU-002-2 v01.00\n'
            '\n'
            'I/O Boards:\n'
            'Board 0 - Model: FP-I/O-3208, Firmware: 01.05, Switches: 32, Drivers: 8\n'
            'Board 1 - Model: FP-I/O-0804, Firmware: 01.05, Switches: 8, Drivers: 4\n'
            'Board 2 - Model: FP-I/O-1616, Firmware: 01.05, Switches: 16, Drivers: 16\n'
            'Board 3 - Model: FP-I/O-1616, Firmware: 01.05, Switches: 16, Drivers: 16\n'
            )

        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def _test_pulse(self):
        self.net_cpu.expected_commands = {
            "TN:04,01": "TN:P"
        }
        # pulse coil 4
        self.machine.coils["c_test"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_long_pulse(self):
        # driver is configured for mode 70, so this should be a regular trigger
        self.net_cpu.expected_commands = {
            "TN:12,01": "TN:P"
        }
        self.machine.coils["c_long_pulse"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_timed_enable(self):
        # enable command
        self.net_cpu.expected_commands = {
            "DN:16,89,00,10,14,FF,C8,88,00": "DN:P"
        }
        self.machine.coils["c_timed_enable"].timed_enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_default_timed_enable(self):
        # enable command
        self.net_cpu.expected_commands = {
            "DN:17,89,00,10,14,FF,C8,88,00": "DN:P"
        }
        self.machine.coils["c_default_timed_enable"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_enable_exception(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.coils["c_test"].enable()
            self.advance_time_and_run(.1)

    def _test_allow_enable(self):
        self.net_cpu.expected_commands = {
            "DN:06,C1,00,18,17,FF,FF,00,00": "DN:P"
        }
        self.machine.coils["c_test_allow_enable"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_pwm_ssm(self):
        self.net_cpu.expected_commands = {
            "DN:13,C1,00,18,0A,FF,88,00,00": "DN:P"
        }
        self.machine.coils["c_hold_ssm"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_nano_reboot(self):
        # NANO reboots
        self.net_cpu.queue.append("!B:00")
        self.advance_time_and_run(.1)
        # assert that MPF will stop
        self.assertTrue(self.machine.stop_future.done())

    def test_rules(self):
        self._test_enable_exception_hw_rule()
        self._test_two_rules_one_switch()
        self._test_hw_rule_pulse()
        self._test_hw_rule_pulse_pwm()
        self._test_hw_rule_pulse_inverted_switch()
        self._test_hw_rule_same_board()

    def _test_enable_exception_hw_rule(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.flippers["f_test_single"].config['main_coil_overwrite']['hold_power'] = 1.0
            self.machine.flippers["f_test_single"].enable()

        self.machine.flippers["f_test_single"].config['main_coil_overwrite']['hold_power'] = None

    def _test_two_rules_one_switch(self):
        self.net_cpu.expected_commands = {
            "TN:04,00,03": "TN:P",
            "TN:06,00,03": "TN:P"
        }
        self.post_event("ac_same_switch")
        self.hit_and_release_switch("s_flipper")
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_hw_rule_pulse(self):
        self.net_cpu.expected_commands = {
            "TN:07,00,16": "TN:P",  # hw rule
        }
        self.machine.autofire_coils["ac_slingshot_test"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "TN:07,02": "TN:P"
        }
        self.machine.autofire_coils["ac_slingshot_test"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_hw_rule_pulse_pwm(self):
        self.net_cpu.expected_commands = {
            "TN:11,01": "TN:P"
        }
        self.machine.coils["c_pulse_pwm_mask"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "DN:11,C1,00,18,0A,AA,92,00,00": "DN:P"
        }
        self.machine.coils["c_pulse_pwm_mask"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_hw_rule_pulse_inverted_switch(self):
        self.net_cpu.expected_commands = {
            "DN:07,11,1A,10,0A,FF,00,00,00": "DN:P",
        }
        self.machine.autofire_coils["ac_inverted_switch"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def _test_hw_rule_same_board(self):
        self.net_cpu.expected_commands = {
            "TN:21,00,07": "DN:P"
        }
        # coil and switch are on different boards but first 8 switches always work
        self.machine.autofire_coils["ac_different_boards"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # switch and coil on board 3. should work
        self.net_cpu.expected_commands = {
            "TN:21,00,39": "DN:P",
        }
        self.machine.autofire_coils["ac_board_3"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # coil and switch are on different boards
        with self.assertRaises(AssertionError):
            self.machine.autofire_coils["ac_broken_combination"].enable()
            self.advance_time_and_run(.1)

    def _switch_hit_cb(self, **kwargs):
        self.switch_hit = True

    def test_switches(self):
        self._test_switch_changes()
        self._test_switch_changes_nc()

    def _test_switch_changes(self):
        self.assertSwitchState("s_flipper", 0)
        self.assertSwitchState("s_flipper_eos", 0)

        self.switch_hit = False
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test", 0)
        self.assertFalse(self.switch_hit)

        self.machine.events.add_handler("s_test_active", self._switch_hit_cb)
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-N:07\r")
        self.advance_time_and_run(1)

        self.assertTrue(self.switch_hit)
        self.assertSwitchState("s_test", 1)
        self.switch_hit = False

        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test", 1)

        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/N:07\r")
        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test", 0)

    def _test_switch_changes_nc(self):
        self.switch_hit = False
        self.machine.events.add_handler("s_test_nc_active", self._switch_hit_cb)
        self.advance_time_and_run(1)
        self.assertSwitchState("s_test_nc", 0)
        self.assertFalse(self.switch_hit)

        self.advance_time_and_run(1)
        self.assertFalse(self.switch_hit)
        self.assertSwitchState("s_test_nc", 0)

        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"-N:1A\r")
        self.advance_time_and_run(1)
        self.assertTrue(self.switch_hit)
        self.assertSwitchState("s_test_nc", 1)

        self.switch_hit = False
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"/N:1A\r")
        self.advance_time_and_run(1)

        self.assertSwitchState("s_test_nc", 0)
        self.assertFalse(self.switch_hit)
        self.switch_hit = False

    def test_flipper_single_coil(self):
        # manual flip no hw rule
        self.net_cpu.expected_commands = {
            "TN:20,01": "TN:P",
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual enable no hw rule
        self.net_cpu.expected_commands = {
            "DN:20,C1,00,18,0A,FF,01,00,00": "DN:P"
        }
        self.machine.coils["c_flipper_main"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual disable no hw rule
        self.net_cpu.expected_commands = {
            "TN:20,02": "TN:P"
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # flipper rule enable
        self.net_cpu.expected_commands = {
            "DN:20,01,01,18,0B,FF,01,00,00": "DN:P",
        }
        self.machine.flippers["f_test_single"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip with hw rule in action
        # TODO I think this is wrong? Check with Dave
        self.net_cpu.expected_commands = {
            "DN:20,09,01,18,0A,FF,01,00,00": "DN:P",    # manual pulse 10ms, trigger 09 = one_shot + driver enable
            "DN:20,01,01,18,0B,FF,01,00,00": "DN:P",    # restore autofire rule
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip with hw rule in action without reconfigure (same pulse)
        self.net_cpu.expected_commands = {
            "TN:20,01": "TN:P",                         # pulse
        }
        self.machine.coils["c_flipper_main"].pulse(11)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual enable with hw rule (same pulse)
        self.net_cpu.expected_commands = {
            "TN:20,03": "TN:P"
        }
        self.machine.coils["c_flipper_main"].enable(pulse_ms=11)
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual disable with hw rule
        self.net_cpu.expected_commands = {
            "TN:20,02": "TN:P",
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual enable with hw rule (different pulse)
        self.net_cpu.expected_commands = {
            "DN:20,C1,01,18,0A,FF,01,00,00": "DN:P",       # configure pwm + enable
        }
        self.machine.coils["c_flipper_main"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual disable with hw rule
        self.net_cpu.expected_commands = {

            "DN:20,01,01,18,0B,FF,01,00,00": "DN:P",    # configure rules
        }
        self.machine.coils["c_flipper_main"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # disable rule
        self.net_cpu.expected_commands = {
            "TN:20,02": "TN:P"                          # disable autofire rule
        }
        self.machine.flippers["f_test_single"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip no hw rule
        self.net_cpu.expected_commands = {
            "DN:20,89,01,18,0A,FF,01,00,00": "DN:P"  # TODO verify this
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        # manual flip again with cached config
        self.net_cpu.expected_commands = {
            "TN:20,01": "TN:P",
        }
        self.machine.coils["c_flipper_main"].pulse()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_flipper_two_coils(self):
        # we pulse the main coil (20)
        # hold coil (21) is pulsed + enabled
        self.net_cpu.expected_commands = {
            "DN:20,01,01,18,0A,FF,00,00,00": "DN:P",
            "DN:21,01,01,18,0A,FF,01,00,00": "DN:P",
        }
        self.machine.flippers["f_test_hold"].enable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

        self.net_cpu.expected_commands = {
            "TN:20,02": "TN:P",
            "TN:21,02": "TN:P"
        }
        self.machine.flippers["f_test_hold"].disable()
        self.advance_time_and_run(.1)
        self.assertFalse(self.net_cpu.expected_commands)

    def test_bootloader_crash(self):
        # Test that the machine stops if the RGB processor sends a bootloader msg
        self.machine.stop = MagicMock()
        self.machine.default_platform.serial_connections['net'].parse_incoming_raw_bytes(b"!B:00\r")
        self.advance_time_and_run(1)
        self.assertTrue(self.machine.stop.called)

    # TODO figure out what to do here
    def test_bootloader_crash_ignored(self):
        # Test that RGB processor bootloader msgs can be ignored
        self.machine.default_platform.config['ignore_reboot'] = True
        self.mock_event('fast_rgb_rebooted')
        self.machine.stop = MagicMock()
        self.machine.default_platform.serial_connections['rgb'].parse_incoming_raw_bytes(b"!B:00\r")
        self.advance_time_and_run(1)
        self.assertFalse(self.machine.stop.called)
        self.assertEventCalled('fast_rgb_rebooted')

    def test_leds(self):

        # Verify the old style number works
        self.assertEqual("00-0", self.machine.lights['old_style_number'].hw_drivers['red'][0].number)

        # Verify 1-based numbers work
        self.assertEqual("FF-0", self.machine.lights['test_max_1_based'].hw_drivers['red'][0].number)

        self.advance_time_and_run()
        device = self.machine.lights["test_led"]   # 0x97
        device2 = self.machine.lights["test_led2"]  #0x98
        self.assertEqual("000000", self.rgb_cpu.leds['97'])
        self.assertEqual("000000", self.rgb_cpu.leds['98'])
        # test led on
        device.on()
        self.advance_time_and_run(1)
        self.assertEqual("FFFFFF", self.rgb_cpu.leds['97'])
        self.assertEqual("000000", self.rgb_cpu.leds['98'])

        device.color("112233")
        device2.color("001122")
        self.advance_time_and_run(1)
        self.assertEqual("112233", self.rgb_cpu.leds['97'])
        self.assertEqual("110022", self.rgb_cpu.leds['98'])  # GRB so ensure it's not 001122

        # test led off
        device.off()
        self.advance_time_and_run(1)
        self.assertEqual("000000", self.rgb_cpu.leds['97'])
        self.assertEqual("110022", self.rgb_cpu.leds['98'])  # GRB so ensure it's not 001122

        # test led color
        device2.color(RGBColor((2, 23, 42)))  #02172A
        self.advance_time_and_run(1)
        self.assertEqual("17022A", self.rgb_cpu.leds['98'])  # GRB so ensure it's not 02172A

        # test led off
        device2.off()
        self.advance_time_and_run(1)
        self.assertEqual("000000", self.rgb_cpu.leds['98'])

        self.advance_time_and_run(.02)

        # fade led over 100ms
        device2.color(RGBColor((100, 100, 100)), fade_ms=100)
        self.advance_time_and_run(.03)
        self.assertTrue(10 < int(self.rgb_cpu.leds['98'][0:2], 16) < 40)
        self.assertTrue(self.rgb_cpu.leds['98'][0:2] == self.rgb_cpu.leds['98'][2:4] == self.rgb_cpu.leds['98'][4:6])
        self.advance_time_and_run(.03)
        self.assertTrue(40 < int(self.rgb_cpu.leds['98'][0:2], 16) < 60)
        self.assertTrue(self.rgb_cpu.leds['98'][0:2] == self.rgb_cpu.leds['98'][2:4] == self.rgb_cpu.leds['98'][4:6])
        self.advance_time_and_run(.03)
        self.assertTrue(60 < int(self.rgb_cpu.leds['98'][0:2], 16) < 90)
        self.assertTrue(self.rgb_cpu.leds['98'][0:2] == self.rgb_cpu.leds['98'][2:4] == self.rgb_cpu.leds['98'][4:6])
        self.advance_time_and_run(2)
        self.assertEqual("646464", self.rgb_cpu.leds['98'])

        # test RGBW LEDs

        rgbw1 = self.machine.lights["test_rgbw1"]
        rgbw2 = self.machine.lights["test_rgbw2"]
        rgbw3 = self.machine.lights["test_rgbw3"]
        rgbw4 = self.machine.lights["test_rgbw4"]
        rgbw5 = self.machine.lights["test_rgbw5"]

        self.assertEqual(rgbw1.hw_drivers['red'][0].number, '21-0')
        self.assertEqual(rgbw1.hw_drivers['green'][0].number, '21-1')
        self.assertEqual(rgbw1.hw_drivers['blue'][0].number, '21-2')
        self.assertEqual(rgbw1.hw_drivers['white'][0].number, '22-0')
        self.assertEqual(rgbw2.hw_drivers['red'][0].number, '22-1')
        self.assertEqual(rgbw2.hw_drivers['green'][0].number, '22-2')
        self.assertEqual(rgbw2.hw_drivers['blue'][0].number, '23-0')
        self.assertEqual(rgbw2.hw_drivers['white'][0].number, '23-1')
        self.assertEqual(rgbw3.hw_drivers['red'][0].number, '23-2')
        self.assertEqual(rgbw3.hw_drivers['green'][0].number, '24-0')
        self.assertEqual(rgbw3.hw_drivers['blue'][0].number, '24-1')
        self.assertEqual(rgbw3.hw_drivers['white'][0].number, '24-2')
        self.assertEqual(rgbw4.hw_drivers['red'][0].number, '25-0')
        self.assertEqual(rgbw4.hw_drivers['green'][0].number, '25-1')
        self.assertEqual(rgbw4.hw_drivers['blue'][0].number, '25-2')
        self.assertEqual(rgbw4.hw_drivers['white'][0].number, '26-0')
        self.assertEqual(rgbw5.hw_drivers['red'][0].number, '28-0')
        self.assertEqual(rgbw5.hw_drivers['green'][0].number, '28-2')
        self.assertEqual(rgbw5.hw_drivers['blue'][0].number, '28-1')
        self.assertEqual(rgbw5.hw_drivers['white'][0].number, '29-2')

        rgbw1.color("ffffff")
        self.advance_time_and_run(1)
        self.assertEqual("000000", self.rgb_cpu.leds['21'])
        self.assertEqual("FF0000", self.rgb_cpu.leds['22'])

        rgbw2.color("445566")
        self.advance_time_and_run(1)
        self.assertEqual("FF0011", self.rgb_cpu.leds['22'])
        self.assertEqual("224400", self.rgb_cpu.leds['23'])

        rgbw5.color("ffffff")
        self.advance_time_and_run(1)
        self.assertEqual("000000", self.rgb_cpu.leds['28'])
        self.assertEqual("0000FF", self.rgb_cpu.leds['29'])

        rgbw5.color("223344") # -> 00112222
        self.advance_time_and_run(1)
        self.assertEqual("002211", self.rgb_cpu.leds['28'])
        self.assertEqual("000022", self.rgb_cpu.leds['29'])


    # TODO need to go through the entire FAST platform and cleanup config errors and error logging in general
    # @expect_startup_error()
    # @test_config("error_lights.yaml")
    # def test_light_errors(self):
    #     self.assertIsInstance(self.startup_error, ConfigFileError)
    #     self.assertEqual(2, self.startup_error.get_error_no())
    #     self.assertEqual("light.test_led", self.startup_error.get_logger_name())
    #     self.assertIsInstance(self.startup_error.__cause__, ConfigFileError)
    #     self.assertEqual(9, self.startup_error.__cause__.get_error_no())
    #     self.assertEqual("FAST", self.startup_error.__cause__.get_logger_name())
    #     self.assertEqual("Light syntax is number-channel (but was \"3\") for light test_led.",
    #                      self.startup_error.__cause__._message)


"""Run regression tests."""

if __name__ == '__main__':
    import os
    import sys
    import unittest

    from mpf.tests.MpfDocTestCase import MpfDocTestCase

    tests = []
    for subdir, dirs, files in os.walk(os.path.join(os.path.dirname(__file__), "mpf", "tests", "regression_tests")):
        for file in files:
            full_path = os.path.join(subdir, file)
            with open(full_path) as f:
                config_string = f.read()
            testcase = MpfDocTestCase(config_string)
            testcase._testMethodDoc = full_path
            tests.append(testcase)

    suite = unittest.TestSuite()
    suite.addTests(tests)
    result = unittest.TextTestRunner(verbosity=1).run(suite)
    if result.wasSuccessful():
        sys.exit(0)
    else:
        sys.exit(1)



"""WS2812 LED on the FAST controller."""
import logging

from typing import Optional
from typing import List

from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface


class FASTRGBLED:

    """FAST RGB LED."""

    __slots__ = ["number", "number_int", "dirty", "log", "channels", "machine", "platform", "hardware_fade_ms"]

    def __init__(self, number: str, platform) -> None:
        """Initialize FAST LED on RGB processor."""
        self.number = number  # 5 char hex string, board address, breakout
        self.number_int = int(number, 16)  # Need this for doing math on the number
        self.dirty = True
        self.machine = platform.machine
        self.platform = platform
        self.hardware_fade_ms = 0
        self.log = logging.getLogger('FASTLED')
        self.channels = [None, None, None]      # type: List[Optional[FASTLEDChannel]]
        # All FAST LEDs are 3 element RGB and are set using hex strings
        self.log.debug("Creating FAST RGB LED at hardware address: %s", self.number)

    def add_channel(self, channel_num: int, channel_obj: "FASTLEDChannel"):
        """Add channel to LED."""
        self.channels[channel_num] = channel_obj

    @property
    def current_color(self):
        """Return current color."""
        result = ""
        self.dirty = False
        current_time = self.machine.clock.get_time()
        for index in [0, 1, 2]:
            channel = self.channels[index]
            if channel:
                brightness, _, done = channel.get_fade_and_brightness(current_time)
                result += f'{int(brightness * 255):02X}'
                if not done:
                    self.dirty = True
            else:
                result += "00"

        return result


class FASTExpLED(FASTRGBLED):

    """FAST RGB LED on an expansion board."""

    __slots__ = ["board_address", "breakout_board", "port",
                 "breakout", "index", "address", "exp_board"]

    def __repr__(self):
        """Return representation of this LED."""
        return f'<FASTExpLED: {self.number}>'

    # pylint: disable-msg=super-init-not-called
    def __init__(self, number: str, hardware_fade_ms: int, platform) -> None:
        """Initialize FAST LED."""
        self.number = number  # 5 char hex string, board address, breakout
        self.number_int = int(number, 16)
        self.platform = platform
        self.address = f'{number[0:3]}'  # '880'
        self.exp_board = platform.exp_boards_by_address[self.address[0:2]]
        self.breakout_board = platform.exp_breakout_boards[self.address]

        # self.port = self.port
        self.dirty = False  # we can reset the board on connection so we don't need to send the first color
        self.machine = platform.machine
        self.platform = platform
        self.hardware_fade_ms = hardware_fade_ms
        self.log = logging.getLogger('FASTLED')
        self.channels = [None, None, None]      # type: List[Optional[FASTLEDChannel]]
        # All FAST LEDs are 3 element RGB and are set using hex strings
        self.log.debug("Creating FAST RGB LED on expansion board at hardware address: %s", self.number)


class FASTLEDChannel(LightPlatformInterface):

    """Represents a single RGB LED channel connected to the FAST hardware platform."""

    __slots__ = ["led", "channel", "_current_fade", "_last_brightness"]

    def __repr__(self):
        """Return representation of this LED."""
        return f'<FASTLEDChannel: {self.led.number}-{self.channel}>'

    def __init__(self, led: FASTRGBLED, channel) -> None:
        """Initialize LED."""
        super().__init__("{}-{}".format(led.number, channel))
        self.led = led
        self.channel = int(channel)
        self._current_fade = (0, -1, 0, -1)
        self._last_brightness = None

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Set brightness via callback."""
        self.led.dirty = True
        self._current_fade = (start_brightness, start_time, target_brightness, target_time)
        self._last_brightness = None

    def get_fade_and_brightness(self, current_time):
        """Return fade + brightness and mark as clean if this is it."""
        if self._last_brightness:
            return self._last_brightness, 0, True
        max_fade_ms = self.led.hardware_fade_ms
        start_brightness, start_time, target_brightness, target_time = self._current_fade
        fade_ms = int((target_time - current_time) * 1000.0)
        if fade_ms > max_fade_ms >= 0:
            fade_ms = max_fade_ms
            ratio = ((current_time + (fade_ms / 1000.0) - start_time) /
                     (target_time - start_time))
            brightness = start_brightness + (target_brightness - start_brightness) * ratio
            done = False
        else:
            if fade_ms < 0:
                fade_ms = 0
            brightness = target_brightness
            self._last_brightness = brightness
            done = True

        # There is a bug that can sometimes cause the start_time to be ahead of the current_time,
        # resulting in a negative brightness value. This may be a floating-point rounding error,
        # or maybe something else. I can't figure it out, so just floor the value to be non-negative.
        if brightness < 0:
            brightness = 0
            self.led.log.warning("Calculated a negative brightness (%s) for led %s channel %s. current_time: %s "
                                 "start_brightness: %s, start_time: %s, target_brightness: %s, target_time: %s",
                                 brightness, self.led, self.channel, current_time, start_brightness, start_time,
                                 target_brightness, target_time)

        return brightness, fade_ms, done

    def get_board_name(self):
        """Return the board of this light."""
        return "FAST LED CPU"

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        return self.led.number_int * 3 + self.channel == other.led.number_int * 3 + other.channel + 1

    def get_successor_number(self):
        """Return next number. We want this in the config format."""
        # if self.channel == 2:
        #     hex_int = int(self.number, 16) + 1  # number is hex string so we need to launder it through an int
        #     return f"{hex_int:X}-0"

        # return f"{self.led.number}-{self.channel + 1}"

        if self.channel == 2:
            return f"{self.led.number_int + 1}-0"

        return f"{self.led.number_int}-{self.channel + 1}"

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return (self.led.number_int, self.channel) < (other.led.number_int, other.channel)



"""Parse config_spec."""
from pkg_resources import iter_entry_points

from mpf.core.utility_functions import Util
from mpf.file_interfaces.yaml_interface import YamlInterface

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.device import Device  # pylint: disable-msg=cyclic-import,unused-import


class ConfigSpecLoader:

    """Config spec loader."""

    @classmethod
    def process_config_spec(cls, spec, path):
        """Process a config spec dictionary."""
        if not isinstance(spec, dict):
            raise AssertionError("Expected a dict at: {} {}".format(path, spec))

        for element, value in spec.items():
            if element.startswith("__"):
                spec[element] = value
            elif isinstance(value, str):
                if value == "ignore":
                    spec[element] = value
                else:
                    spec[element] = value.split('|')
                    if len(spec[element]) != 3:
                        raise AssertionError("Format incorrect: {}".format(value))
            else:
                spec[element] = cls.process_config_spec(value, path + ":" + element)

        return spec

    @staticmethod
    def load_external_platform_config_specs(config):
        """Load config spec for external platforms."""
        for platform_entry in iter_entry_points(group='mpf.platforms'):
            config_spec = platform_entry.load().get_config_spec()

            if config_spec:
                # add specific config spec if platform has any
                config[config_spec[1]] =\
                    ConfigSpecLoader.process_config_spec(YamlInterface.process(config_spec[0]), config_spec[1])
        return config

    @staticmethod
    def load_device_config_specs(config_spec, machine_config):
        """Load device config specs."""
        for device_type in machine_config['mpf']['device_modules'].values():
            device_cls = Util.string_to_class(device_type)      # type: Device
            if device_cls.get_config_spec():
                # add specific config spec if device has any
                config_spec[device_cls.config_section] = ConfigSpecLoader.process_config_spec(
                    YamlInterface.process(device_cls.get_config_spec()),
                    device_cls.config_section)

        return config_spec



"""Specialized text support classes for segment displays."""
import abc
from collections import namedtuple
from typing import Optional, List, Union

from mpf.core.rgb_color import RGBColor

DisplayCharacter = namedtuple("DisplayCharacter", ["char_code", "dot", "comma", "color"])

DOT_CODE = ord(".")
COMMA_CODE = ord(",")
SPACE_CODE = ord(" ")


class SegmentDisplayText(metaclass=abc.ABCMeta):

    """A list of characters with specialized functions for segment displays. Use for display text effects."""

    __slots__ = ["_text", "embed_dots", "embed_commas", "use_dots_for_commas"]

    def __init__(self, char_list, embed_dots, embed_commas, use_dots_for_commas):
        """Initialize segment display text."""
        self.embed_dots = embed_dots
        self.embed_commas = embed_commas
        self.use_dots_for_commas = use_dots_for_commas
        self._text = char_list

    # pylint: disable=too-many-arguments
    @classmethod
    def from_str_with_color(cls, text: str, display_size: int, collapse_dots: bool, collapse_commas: bool,
                            use_dots_for_commas: bool, colors: List[RGBColor]) -> "ColoredSegmentDisplayText":
        """Create colored text."""
        return ColoredSegmentDisplayText(
            cls._create_characters(text, display_size, collapse_dots, collapse_commas, use_dots_for_commas, colors[:]),
            collapse_dots, collapse_commas, use_dots_for_commas)

    # pylint: disable=too-many-arguments
    @classmethod
    def from_str(cls, text: str, display_size: int, collapse_dots: bool, collapse_commas: bool,
                 use_dots_for_commas: bool, colors: Optional[List[RGBColor]] = None) -> \
            Union["ColoredSegmentDisplayText", "UncoloredSegmentDisplayText"]:
        """Create from string."""
        if colors:
            return cls.from_str_with_color(text, display_size, collapse_dots, collapse_commas,
                                           use_dots_for_commas, colors[:])

        char_colors = [None] * len(text)
        return UncoloredSegmentDisplayText(
            cls._create_characters(text, display_size, collapse_dots, collapse_commas, use_dots_for_commas,
                                   char_colors), collapse_dots, collapse_commas, use_dots_for_commas)

    @classmethod
    def _embed_dots_and_commas(cls, text: str, collapse_dots: bool, collapse_commas: bool, use_dots_for_commas: bool):
        """Return text with embedded dots and commas."""
        char_has_dot = False
        char_has_comma = False
        char_list = []
        for char in reversed(text):
            char_code = ord(char)
            if collapse_dots and char_code == DOT_CODE or use_dots_for_commas and char_code == COMMA_CODE:
                char_has_dot = True
                continue
            if collapse_commas and char_code == COMMA_CODE:
                char_has_comma = True
                continue
            char_list.insert(0, (char_code, char_has_dot, char_has_comma))
            char_has_dot = False
            char_has_comma = False

        return char_list

    # pylint: disable-msg=too-many-locals
    @classmethod
    def _create_characters(cls, text: str, display_size: int, collapse_dots: bool, collapse_commas: bool,
                           use_dots_for_commas: bool, colors: List[Optional[RGBColor]]) -> List[DisplayCharacter]:
        """Create characters from text and color them.

        - Colors are used from the left to the right (starting with the first character).
        - If colors are shorter than text the last color is repeated for text.
        - The first color is used to pad the text to the left if text is shorter than the display - thus text is right
          aligned.
        - Dots and commas are embedded on the fly.
        """
        char_list = []
        left_pad_color = colors[0] if colors else None
        default_right_color = colors[len(colors) - 1] if colors else None
        uncolored_chars = cls._embed_dots_and_commas(text, collapse_dots, collapse_commas, use_dots_for_commas)
        colors = colors[-len(uncolored_chars):]
        for char_code, char_has_dot, char_has_comma in uncolored_chars:
            color = colors.pop(0) if colors else default_right_color
            char_list.append(DisplayCharacter(char_code, char_has_dot, char_has_comma, color))

        # ensure list is the same size as the segment display (cut off on left or right justify characters)
        current_length = len(char_list)
        if current_length > display_size:
            for _ in range(current_length - display_size):
                char_list.pop(0)
        elif current_length < display_size:
            for _ in range(display_size - current_length):
                char_list.insert(0, DisplayCharacter(SPACE_CODE, False, False, left_pad_color))

        return char_list

    def blank_segments(self, flash_mask) -> "SegmentDisplayText":
        """Return new SegmentDisplayText with chars blanked."""
        return ColoredSegmentDisplayText(
            [char if mask != "F" else DisplayCharacter(SPACE_CODE, False, False, char.color)
             for char, mask in zip(self._text, flash_mask)],
            self.embed_dots, self.embed_commas, self.use_dots_for_commas
        )

    def convert_to_str(self):
        """Convert back to normal text string."""
        text = ""
        for display_character in self:
            text += chr(display_character.char_code)
            if display_character.dot:
                text += "."
            if display_character.comma:
                text += ","

        return text

    def __len__(self):
        """Return length."""
        return self._text.__len__()

    def __getitem__(self, item):
        """Return item or slice."""
        if isinstance(item, slice):
            return self.__class__(self._text.__getitem__(item), self.embed_dots, self.embed_commas,
                                  self.use_dots_for_commas)

        return self._text.__getitem__(item)

    def __eq__(self, other):
        """Return true if two texts and colors are the same."""
        # pylint: disable-msg=protected-access
        return isinstance(other, SegmentDisplayText) and self._text == other._text

    def extend(self, other_list):
        """Extend list."""
        # pylint: disable-msg=protected-access
        self._text.extend(other_list._text)

    @abc.abstractmethod
    def get_colors(self) -> Optional[List[RGBColor]]:
        """Get the list of colors for each character (if set)."""
        raise NotImplementedError()

    def __repr__(self):
        """Return string representation."""
        return "<{} {}>".format(self.__class__, self._text)


class UncoloredSegmentDisplayText(SegmentDisplayText):

    """Segment text without colors."""

    def get_colors(self) -> None:
        """Return None as we are transparent."""
        return None


class ColoredSegmentDisplayText(SegmentDisplayText):

    """Segment text with colors."""

    def get_colors(self) -> List[RGBColor]:
        """Get the list of colors for each character (if set)."""
        return [display_character.color for display_character in self]



"""Test high score mode."""
from unittest.mock import MagicMock
from mpf.tests.MpfBcpTestCase import MpfBcpTestCase


class TestHighScoreMode(MpfBcpTestCase):

    def get_config_file(self):
        return 'high_score.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/high_score_vars/'

    def start_game(self, num_players=1):
        self._bcp_client.send = MagicMock()

        self.machine.playfield.add_ball = MagicMock()
        self.machine.events.post('game_start')
        self.advance_time_and_run()
        self.machine.game.balls_in_play = 1
        self.assertIsNotNone(self.machine.game)

        while self.machine.game.num_players < num_players:
            self.machine.game.request_player_add()
            self.advance_time_and_run()

    def test_high_score_one_var(self):
        self.advance_time_and_run()
        self.mock_event("high_score_enter_initials")
        self.mock_event("loops_award_display")
        # tests loop award with additional variables
        self.machine.modes["high_score"].high_scores = dict()
        self.machine.modes["high_score"].high_scores['loops'] = [('BIL', 2)]

        self.start_game(1)
        self.machine.game.player_list[0].loops = 50
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        # Award initials
        self.assertEqual(1, self._events['high_score_enter_initials'])
        self._bcp_client.send.reset_mock()

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='NEW')))
        self.advance_time_and_run(.5)

        self.assertEventCalledWith("loops_award_display", award='LOOP CHAMP', player_name='NEW', value=50,
                                   player_num=1, category_name='loops')
        self.advance_time_and_run(4)

        # High score done
        self.assertFalse(self.machine.modes["high_score"].active)

        # verify the data is accurate

        new_score_data = {'score': [], 'loops': [['NEW', 50, {'player_number': 1}]], 'hits': []}

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_high_score_multiple_vars(self):
        self.advance_time_and_run()
        self.mock_event("high_score_enter_initials")
        self.mock_event("hits_award_display")
        # tests loop award with additional variables
        self.machine.modes["high_score"].high_scores = dict()
        self.machine.modes["high_score"].high_scores['hits'] = [['AAA', 2]]

        self.start_game(1)
        self.machine.game.player_list[0].hits = 50
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        # Award initials
        self.assertEqual(1, self._events['high_score_enter_initials'])
        self._bcp_client.send.reset_mock()

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='NEW')))
        self.advance_time_and_run(.5)

        self.assertEventCalledWith("hits_award_display", award='MOST HITS', player_name='NEW', value=50,
                                   player_num=1, category_name='hits')
        self.advance_time_and_run(4)

        # High score done
        self.assertFalse(self.machine.modes["high_score"].active)

        # verify the data is accurate

        new_score_data = {'score': [],
                          'loops': [],
                          'hits': [['NEW', 50, {'player_number': 1, 'machine_credits_string': 'FREE PLAY'}]]}

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)



# pylint: disable-msg=too-many-lines
"""Contains the MachineController base class."""
import asyncio
import logging
import sys
import threading
from typing import Any, Callable, Dict, List, Set, Optional

from pkg_resources import iter_entry_points

from mpf._version import __version__
from mpf.core.clock import ClockBase
from mpf.core.config_validator import ConfigValidator
from mpf.core.crash_reporter import report_crash
from mpf.core.data_manager import DataManager
from mpf.core.delays import DelayManager
from mpf.core.device_manager import DeviceCollection
from mpf.core.logging import LogMixin
from mpf.core.machine_vars import MachineVariables
from mpf.core.utility_functions import Util
from mpf.core.config_loader import MpfConfig
from mpf.core.plugin import MpfPlugin

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.modes.game.code.game import Game   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.events import EventManager    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.switch_controller import SwitchController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.show_controller import ShowController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.service_controller import ServiceController   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.light_controller import LightController   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.platform_controller import PlatformController     # pylint: disable-msg=cyclic-import,unused-import

    from mpf.core.custom_code import CustomCode     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.mode_controller import ModeController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.settings_controller import SettingsController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.bcp.bcp import Bcp    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.text_ui import TextUi     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.assets.show import Show    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.switch import Switch   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.driver import Driver   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.mode import Mode  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_device.ball_device import BallDevice  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.ball_controller import BallController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.playfield import Playfield     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.placeholder_manager import PlaceholderManager     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.platforms.smart_virtual import SmartVirtualHardwarePlatform    # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.core.device_manager import DeviceManager   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.plugins.auditor import Auditor     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.light import Light     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.accelerometer import Accelerometer     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.drop_target import DropTarget, DropTargetBank  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.logic_blocks import Accrual, Sequence, Counter     # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.devices.servo import Servo     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.segment_display.segment_display import \
        SegmentDisplay  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.shot_group import ShotGroup    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.shot import Shot   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.motor import Motor     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.digital_output import DigitalOutput    # pylint: disable-msg=cyclic-import,unused-import
    from logging import Logger  # noqa
    from mpf.devices.autofire import AutofireCoil   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.stepper import Stepper     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.config_players.show_player import ShowPlayer   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.dmd import Dmd     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.rgb_dmd import RgbDmd  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.flipper import Flipper     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.diverter import Diverter   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.multiball_lock import MultiballLock    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.multiball import Multiball     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_hold import BallHold      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_save import BallSave      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.magnet import Magnet           # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.state_machine import StateMachine  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.extra_ball import ExtraBall    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.extra_ball_group import ExtraBallGroup     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.achievement_group import AchievementGroup  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.achievement import Achievement     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.combo_switch import ComboSwitch    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.score_queue import ScoreQueue      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.spinner import Spinner      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.blinkenlight import Blinkenlight   # pylint: disable-msg=cyclic-import,unused-import


# pylint: disable-msg=too-many-instance-attributes
class MachineController(LogMixin):

    """Base class for the Machine Controller object.

    The machine controller is the main entity of the entire framework. It's the
    main part that's in charge and makes things happen.

    Args:
    ----
        config(MpfConfig): The machine configuration
        options(dict): A dictionary of options built from the command line options
            used to launch mpf.py.
    """

    __slots__ = ["options", "mpf_path", "machine_path", "_exception", "_boot_holds",
                 "is_init_done", "_done", "monitors", "plugins", "custom_code", "modes", "game",
                 "variables", "thread_stopper", "config", "config_validator",
                 "delay", "hardware_platforms", "default_platform", "clock",
                 "stop_future", "events", "switch_controller", "mode_controller", "settings",
                 "bcp", "ball_controller", "show_controller", "placeholder_manager", "device_manager", "auditor",
                 "tui", "service", "switches", "shows", "coils", "ball_devices", "lights", "playfield", "playfields",
                 "autofire_coils", "_crash_handlers", "__dict__", "mpf_config", "is_shutting_down"]

    # pylint: disable-msg=too-many-statements
    def __init__(self, options: dict, config: MpfConfig) -> None:
        """Initialize machine controller."""
        super().__init__()
        self.log = logging.getLogger("Machine")     # type: Logger
        self.log.info("Mission Pinball Framework Core Engine v%s", __version__)
        self._crash_handlers = []   # type: List[Callable]
        self.is_shutting_down = False

        self.log.info("Command line arguments: %s", options)
        self.options = options

        self.mpf_path = config.get_mpf_path()
        self.log.info("MPF path: %s", self.mpf_path)

        self.machine_path = config.get_machine_path()
        self.log.info("Machine path: %s", self.machine_path)

        self.verify_system_info()
        self._exception = None      # type: Any
        self._boot_holds = set()    # type: Set[str]
        self.is_init_done = None    # type: Optional[asyncio.Event]

        self._done = False
        self.monitors = dict()      # type: Dict[str, Set[Callable]]
        self.plugins = list()       # type: List[Any]
        self.custom_code = list()   # type: List[CustomCode]
        self.modes = DeviceCollection(self, 'modes', None)          # type: Dict[str, Mode]
        self.game = None            # type: Optional[Game]
        self.thread_stopper = threading.Event()

        self.config = config.get_machine_config()       # type: Any
        self.mpf_config = config                        # type: MpfConfig
        self.config_validator = ConfigValidator(self, self.mpf_config.get_config_spec())

        self.variables = MachineVariables(self)  # type: MachineVariables

        # add some type hints
        if MYPY:   # pragma: no cover
            # controllers
            self.events = self.events                               # type: EventManager
            self.switch_controller = self.switch_controller         # type: SwitchController
            self.mode_controller = self.mode_controller             # type: ModeController
            self.settings = self.settings                           # type: SettingsController
            self.bcp = self.bcp                                     # type: Bcp
            self.ball_controller = self.ball_controller             # type: BallController
            self.show_controller = self.show_controller             # type: ShowController
            self.placeholder_manager = self.placeholder_manager     # type: PlaceholderManager
            self.device_manager = self.device_manager               # type: DeviceManager
            self.auditor = self.auditor                             # type: Auditor
            self.tui = self.tui                                     # type: TextUi
            self.service = self.service                             # type: ServiceController
            self.show_player = self.show_player                     # type: ShowPlayer
            self.light_controller = self.light_controller           # type: LightController
            self.platform_controller = self.platform_controller     # type: PlatformController

            # devices
            self.autofire_coils = {}                    # type: Dict[str, AutofireCoil]
            self.motors = {}                            # type: Dict[str, Motor]
            self.digital_outputs = {}                   # type: Dict[str, DigitalOutput]
            self.shows = {}                             # type: Dict[str, Show]
            self.shots = {}                             # type: Dict[str, Shot]
            self.shot_groups = {}                       # type: Dict[str, ShotGroup]
            self.switches = {}                          # type: Dict[str, Switch]
            self.steppers = {}                          # type: Dict[str, Stepper]
            self.coils = {}                             # type: Dict[str, Driver]
            self.lights = {}                            # type: Dict[str, Light]
            self.ball_devices = {}                      # type: Dict[str, BallDevice]
            self.accelerometers = {}                    # type: Dict[str, Accelerometer]
            self.playfield = None                       # type: Optional[Playfield]
            self.playfields = {}                        # type: Dict[str, Playfield]
            self.counters = {}                          # type: Dict[str, Counter]
            self.sequences = {}                         # type: Dict[str, Sequence]
            self.accruals = {}                          # type: Dict[str, Accrual]
            self.drop_targets = {}                      # type: Dict[str, DropTarget]
            self.drop_target_banks = {}                 # type: Dict[str, DropTargetBank]
            self.servos = {}                            # type: Dict[str, Servo]
            self.segment_displays = {}                  # type: Dict[str, SegmentDisplay]
            self.dmds = {}                              # type: Dict[str, Dmd]
            self.rgb_dmds = {}                          # type: Dict[str, RgbDmd]
            self.flippers = {}                          # type: Dict[str, Flipper]
            self.diverters = {}                         # type: Dict[str, Diverter]
            self.multiball_locks = {}                   # type: Dict[str, MultiballLock]
            self.multiballs = {}                        # type: Dict[str, Multiball]
            self.ball_holds = {}                        # type: Dict[str, BallHold]
            self.ball_saves = {}                        # type: Dict[str, BallSave]
            self.magnets = {}                           # type: Dict[str, Magnet]
            self.state_machines = {}                    # type: Dict[str, StateMachine]
            self.extra_balls = {}                       # type: Dict[str, ExtraBall]
            self.extra_ball_groups = {}                 # type: Dict[str, ExtraBallGroup]
            self.achievements = {}                      # type: Dict[str, Achievement]
            self.achievement_groups = {}                # type: Dict[str, AchievementGroup]
            self.combo_switches = {}                    # type: Dict[str, ComboSwitch]
            self.score_queues = {}                      # type: Dict[str, ScoreQueue]
            self.spinners = {}                          # type: Dict[str, Spinner]
            self.blinkenlights = {}                     # type: Dict[str, Blinkenlight]

        self._set_machine_path()

        self.configure_logging(
            'Machine',
            self.config['logging']['console']['machine_controller'],
            self.config['logging']['file']['machine_controller'])

        self.delay = DelayManager(self)

        self.hardware_platforms = dict()    # type: Dict[str, SmartVirtualHardwarePlatform]
        self.default_platform = None        # type: Optional[SmartVirtualHardwarePlatform]

        self.clock = self._load_clock()
        self.stop_future = asyncio.Future()     # type: asyncio.Future

    def add_crash_handler(self, handler: Callable):
        """Add a crash handler which is called on a crash.

        This can be used to restore the output and prepare logging.
        """
        self._crash_handlers.append(handler)

    async def initialize_core_and_hardware(self) -> None:
        """Load core modules and hardware."""
        self._boot_holds = set()
        self.is_init_done = asyncio.Event()
        self.register_boot_hold('init')
        self._load_hardware_platforms()

        self._load_core_modules()
        # order is specified in mpfconfig.yaml

        self._validate_config()

        # This is called so hw platforms have a chance to register for events,
        # and/or anything else they need to do with core modules since
        # they're not set up yet when the hw platforms are constructed.
        await self._initialize_platforms()

    async def initialize(self) -> None:
        """Initialize machine."""
        await self.initialize_core_and_hardware()

        self._initialize_credit_string()

        self._register_config_players()
        self._register_system_events()
        self._load_machine_vars()
        await self._run_init_phases()
        self._init_phases_complete()

        await self._start_platforms()

        # wait until all boot holds were released
        assert self.is_init_done is not None
        await self.is_init_done.wait()
        await self.init_done()

    def _exception_handler(self, loop, context):    # pragma: no cover
        """Handle asyncio loop exceptions."""
        # call original exception handler
        loop.set_exception_handler(None)
        loop.call_exception_handler(context)

        # remember exception
        self._exception = context
        self.stop("Exception thrown")

    def _load_clock(self) -> ClockBase:  # pragma: no cover
        """Load clock and loop."""
        clock = ClockBase(self)
        clock.loop.set_exception_handler(self._exception_handler)
        return clock

    async def _run_init_phases(self) -> None:
        """Run init phases."""
        await self.events.post_queue_async("init_phase_1")
        '''event: init_phase_1

        desc: Posted during the initial boot up of MPF.
        '''
        await self.events.post_queue_async("init_phase_2")
        '''event: init_phase_2

        desc: Posted during the initial boot up of MPF.
        '''
        self._load_plugins()
        await self.events.post_queue_async("init_phase_3")
        '''event: init_phase_3

        desc: Posted during the initial boot up of MPF.
        '''
        self._load_custom_code()

        await self.events.post_queue_async("init_phase_4")
        '''event: init_phase_4

        desc: Posted during the initial boot up of MPF.
        '''

        await self.events.post_queue_async("init_phase_5")
        '''event: init_phase_5

        desc: Posted during the initial boot up of MPF.
        '''

    def _init_phases_complete(self, **kwargs) -> None:
        """Cleanup after init and remove boot holds."""
        del kwargs
        self.events.remove_all_handlers_for_event("init_phase_1")
        self.events.remove_all_handlers_for_event("init_phase_2")
        self.events.remove_all_handlers_for_event("init_phase_3")
        self.events.remove_all_handlers_for_event("init_phase_4")
        self.events.remove_all_handlers_for_event("init_phase_5")

        self.clear_boot_hold('init')

    async def _initialize_platforms(self) -> None:
        """Initialize all used hardware platforms."""
        init_done = []
        # collect all platform init futures
        for hardware_platform in list(self.hardware_platforms.values()):
            init_done.append(hardware_platform.initialize())

        # wait for all of them in parallel
        results = await asyncio.wait([asyncio.create_task(init_done) for init_done in init_done])
        for result in results[0]:
            result.result()

    async def _start_platforms(self) -> None:
        """Start all used hardware platforms."""
        for hardware_platform in list(self.hardware_platforms.values()):
            await hardware_platform.start()
            if not hardware_platform.features['tickless']:
                self.clock.schedule_interval(hardware_platform.tick, 1 / self.config['mpf']['default_platform_hz'])

    def _initialize_credit_string(self):
        """Set default credit string."""
        # Do this here so there's a credit_string var even if they're not using
        # the credits mode
        try:
            credit_string = self.config['credits']['free_play_string']
        except KeyError:
            credit_string = 'FREE PLAY'

        self.variables.set_machine_var('credits_string', credit_string)
        '''machine_var: credits_string

        desc: Holds a displayable string which shows how many
        credits are on the machine. For example, "CREDITS: 1". If the machine
        is set to free play, the value of this string will be "FREE PLAY".

        You can change the format and value of this string in the ``credits:``
        section of the machine config file.
        '''

    def wait_for_stop(self):
        """Wait until machine stops."""
        return asyncio.shield(self.stop_future)

    def _validate_config(self) -> None:
        """Validate game and machine config."""
        self.validate_machine_config_section('machine')
        self.validate_machine_config_section('game')
        self.validate_machine_config_section('mpf')

    def validate_machine_config_section(self, section: str) -> None:
        """Validate a config section."""
        if section not in self.config_validator.get_config_spec():
            return

        if section not in self.config:
            self.config[section] = dict()

        self.config[section] = self.config_validator.validate_config(
            section, self.config[section], section)

    def _register_system_events(self) -> None:
        """Register default event handlers."""
        self.events.add_handler('quit', self.stop)
        self.events.add_handler(self.config['mpf']['switch_tag_event'].
                                replace('%', 'quit'), self.stop)

    def _register_config_players(self) -> None:
        """Register config players."""
        # todo move this to config_player module
        for name, module_class in self.config['mpf']['config_players'].items():
            config_player_class = Util.string_to_class(module_class)
            setattr(self, '{}_player'.format(name),
                    config_player_class(self))

        self._register_plugin_config_players()

    def _register_plugin_config_players(self):
        """Register plugin config players."""
        self.debug_log("Registering Plugin Config Players")
        for entry_point in iter_entry_points(group='mpf.config_player',
                                             name=None):
            self.debug_log("Registering %s", entry_point)
            name, player = entry_point.load()(self)
            setattr(self, '{}_player'.format(name), player)

    def create_data_manager(self, config_name: str) -> DataManager:     # pragma: no cover
        """Return a new DataManager for a certain config.

        Args:
        ----
            config_name: Name of the config
        """
        return DataManager(self, config_name)

    def _load_machine_vars(self) -> None:
        """Load machine vars from data manager."""
        machine_var_data_manager = self.create_data_manager('machine_vars')
        current_time = self.clock.get_datetime().timestamp()

        self.variables.load_machine_vars(machine_var_data_manager, current_time)

    def _set_machine_path(self) -> None:
        """Add the machine folder to sys.path so we can import modules from it."""
        sys.path.insert(0, self.machine_path)

    def verify_system_info(self):
        """Dump information about the Python installation to the log.

        Information includes Python version, Python executable, platform, and
        core architecture.
        """
        python_version_info = sys.version_info

        # if not (python_version_info[0] == 3 and python_version_info[1] in (5, 6, 7, 8, 9)):
        #     raise AssertionError("Incorrect Python version. MPF requires "
        #                          "Python 3.5, 3.6, 3.7, 3.8 or 3.9. You have Python {}.{}.{}."
        #                          .format(python_version_info[0], python_version_info[1],
        #                                  python_version_info[2]))

        self.log.info("Platform: %s", sys.platform)
        self.log.info("Python executable location: %s", sys.executable)

        if sys.maxsize < 2**32:
            self.log.info("Python version: %s.%s.%s (32-bit)", python_version_info[0],
                          python_version_info[1], python_version_info[2])
        else:
            self.log.info("Python version: %s.%s.%s (64-bit)", python_version_info[0],
                          python_version_info[1], python_version_info[2])

    def _load_core_modules(self) -> None:
        """Load core modules."""
        self.debug_log("Loading core modules...")
        for name, module_class in self.config['mpf']['core_modules'].items():
            self.debug_log("Loading '%s' core module", module_class)
            m = Util.string_to_class(module_class)(self)
            setattr(self, name, m)

    def _load_hardware_platforms(self) -> None:
        """Load all hardware platforms."""
        self.validate_machine_config_section('hardware')

        # load internal platforms
        self.add_platform("drivers")

        # if platform is forced use that one
        if self.options['force_platform']:
            self.add_platform(self.options['force_platform'])
            self.set_default_platform(self.options['force_platform'])
            return

        # otherwise load all platforms
        for section, platforms in self.config['hardware'].items():
            if section == 'driverboards':
                continue
            for hardware_platform in platforms:
                if hardware_platform.lower() != 'default':
                    self.add_platform(hardware_platform)

        # set default platform
        self.set_default_platform(self.config['hardware']['platform'][0])

    def _load_plugins(self) -> None:
        """Load plugins."""
        self.debug_log("Loading plugins...")

        for plugin in self.config['mpf']['plugins']:

            self.debug_log("Loading '%s' plugin", plugin)

            plugin_obj = Util.string_to_class(plugin)(self)  # type: MpfPlugin
            if plugin_obj.is_plugin_enabled:
                self.debug_log("Including plugin %s", plugin_obj.name)
                plugin_obj.initialize()
                self.plugins.append(plugin_obj)
            else:
                self.debug_log("Excluding plugin %s because it's not enabled", plugin_obj.name)

    def _load_custom_code(self) -> None:
        """Load custom code."""
        if 'custom_code' in self.config:
            self.debug_log("Loading custom code.")

            for custom_code in Util.string_to_event_list(self.config['custom_code']):

                self.debug_log("Loading %s custom code", custom_code)

                custom_code_obj = Util.string_to_class(custom_code)(
                    machine=self,
                    name=custom_code)

                self.custom_code.append(custom_code_obj)

    async def reset(self) -> None:
        """Reset the machine.

        This method is safe to call. It essentially sets up everything from
        scratch without reloading the config files and assets from disk. This
        method is called after a game ends and before attract mode begins.
        """
        self.debug_log('Resetting...')

        await self.events.post_queue_async('machine_reset_phase_1')
        '''Event: machine_reset_phase_1

        Desc: The first phase of resetting the machine.

        These events are posted when MPF boots (after the init_phase events are
        posted), and they're also posted subsequently when the machine is reset
        (after exiting the service mode, for example).

        This is an async queue event. The machine reset phase 1 will not be complete
        until the queue is cleared.

        '''

        await self.events.post_queue_async('machine_reset_phase_2')
        '''Event: machine_reset_phase_2

        Desc: The second phase of resetting the machine.

        These events are posted when MPF boots (after the init_phase events are
        posted), and they're also posted subsequently when the machine is reset
        (after exiting the service mode, for example).

        This is an async queue event. The machine reset phase 2 will not be complete
        until the queue is cleared.

        '''

        await self.events.post_queue_async('machine_reset_phase_3')
        '''Event: machine_reset_phase_3

        Desc: The third phase of resetting the machine.

        These events are posted when MPF boots (after the init_phase events are
        posted), and they're also posted subsequently when the machine is reset
        (after exiting the service mode, for example).

        This is an async queue event. The machine reset phase 3 will not be complete
        until the queue is cleared.

        '''

        """Called when the machine reset process is complete."""
        self.debug_log('Reset Complete')
        await self.events.post_async('reset_complete')
        '''event: reset_complete

        desc: The machine reset process is complete

        '''

    def add_platform(self, name: str) -> None:
        """Make an additional hardware platform interface available to MPF.

        Args:
        ----
            name: String name of the platform to add. Must match the name of a
                platform file in the mpf/platforms folder (without the .py
                extension).
        """
        if name not in self.hardware_platforms:
            if name in self.config['mpf']['platforms']:
                # if platform is in config load it
                try:
                    hardware_platform = Util.string_to_class(self.config['mpf']['platforms'][name])
                except ImportError as e:  # pragma: no cover
                    if e.name != name:  # do not swallow unrelated errors
                        raise
                    raise ImportError("Cannot add hardware platform {}. This is "
                                      "not a valid platform name".format(name))

            else:
                # check entry points
                entry_points = list(iter_entry_points(group='mpf.platforms', name=name))
                if entry_points:
                    # load platform from entry point
                    self.debug_log("Loading platform %s from external entry_point", name)
                    if len(entry_points) != 1:
                        raise AssertionError("Got more than one entry point for platform {}: {}".format(name,
                                                                                                        entry_points))

                    hardware_platform = entry_points[0].load()
                else:
                    raise AssertionError("Unknown platform {}".format(name))

            self.hardware_platforms[name] = hardware_platform(self)

    def set_default_platform(self, name: str) -> None:
        """Set the default platform.

        It is used if a device class-specific or device-specific platform is not specified.

        Args:
        ----
            name: String name of the platform to set to default.
        """
        try:
            self.default_platform = self.hardware_platforms[name]
            self.debug_log("Setting default platform to '%s'", name)
        except KeyError:
            raise AssertionError("Cannot set default platform to '{}', as that's not"
                                 " a currently active platform".format(name))

    def register_monitor(self, monitor_class: str, monitor: Callable[..., Any]) -> None:
        """Register a monitor.

        Args:
        ----
            monitor_class: String name of the monitor class for this monitor
                that's being registered.
            monitor: Callback to notify

        MPF uses monitors to allow components to monitor certain internal
        elements of MPF.

        For example, a player variable monitor could be setup to be notified of
        any changes to a player variable, or a switch monitor could be used to
        allow a plugin to be notified of any changes to any switches.

        The MachineController's list of registered monitors doesn't actually
        do anything. Rather it's a dictionary of sets which the monitors
        themselves can reference when they need to do something. We just needed
        a central registry of monitors.

        """
        if monitor_class not in self.monitors:
            self.monitors[monitor_class] = set()

        self.monitors[monitor_class].add(monitor)

    def initialize_mpf(self):
        """Initialize MPF."""
        self.info_log("Initializing MPF...")
        timeout = 30 if self.options["production"] else None
        try:
            init = asyncio.ensure_future(self.initialize())
            self.clock.loop.run_until_complete(Util.first([init, self.stop_future], cancel_others=False,
                                                          timeout=timeout))
        except asyncio.TimeoutError:
            self._crash_shutdown()
            self.error_log("MPF needed more than %ss for initialization. Aborting!", timeout)
            return False
        except RuntimeError as e:
            self._crash_shutdown()
            # do not show a runtime useless runtime error
            self.error_log("Failed to initialize MPF")
            report_crash(e, "init_runtime_error", self.config)
            return False
        if init.done() and init.exception():
            self._crash_shutdown()
            try:
                raise init.exception()
            except:     # noqa
                self.log.exception("Failed to initialize MPF")
                report_crash(init.exception(), "init_exception", self.config)
            return False

        return True

    def run(self) -> None:
        """Start the main machine run loop."""
        if not self.initialize_mpf():
            return

        self.info_log("Starting the main run loop.")
        self._run_loop()

    def stop_with_exception(self, exception) -> None:
        """Stop MPF because of an exception.

        This may be called from other threads.
        """
        self._exception = exception
        self.clock.loop.call_soon_threadsafe(self._stop_loop, "Exception")

    def stop(self, reason=None, **kwargs) -> None:
        """Perform a graceful exit of MPF."""
        del kwargs
        if self.stop_future.done():
            return

        if self.clock and self.clock.loop and not self.clock.loop.is_closed():
            self.clock.loop.call_soon_threadsafe(self._stop_loop, reason)
        self.is_shutting_down = True

    def _stop_loop(self, reason):
        if self.stop_future.done():
            return
        self.stop_future.set_result(reason)

    def _do_stop(self) -> None:
        self.is_shutting_down = True
        self.log.info("Shutting down...")
        self.events.post('shutdown')
        '''event: shutdown
        desc: Posted when the machine is shutting down to give all modules a
        chance to shut down gracefully.

        '''

        self.events.process_event_queue()
        self.shutdown()

    def _crash_shutdown(self):
        """MPF crashed. Cleanup as good as we can."""
        # call crash handlers
        for handler in self._crash_handlers:
            handler()
        if hasattr(self, "events") and hasattr(self, "clock"):
            # if we already got events and a clock use normal shutdown
            self._do_stop()
        else:
            # otherwise just shutdown
            self.shutdown()

    @staticmethod
    async def _stop_tasks(tasks):
        open_tasks = [t for t in tasks if not t.done()]
        if not open_tasks:
            return

        await asyncio.wait(open_tasks, timeout=1.0)

    def shutdown(self) -> None:
        """Shutdown the machine."""
        self.is_shutting_down = True
        if not self.stop_future.done():
            self.stop_future.set_result("Graceful shutdown")
        self.thread_stopper.set()
        if hasattr(self, "device_manager"):
            self.device_manager.stop_devices()
        self._platform_stop()

        # this is needed to properly close all sockets
        try:
            if hasattr(asyncio, "all_tasks"):
                self.clock.loop.run_until_complete(self._stop_tasks(asyncio.all_tasks(loop=self.clock.loop)))
            else:
                self.clock.loop.run_until_complete(self._stop_tasks(asyncio.Task.all_tasks(loop=self.clock.loop)))
        except RuntimeError:
            print("Failed to stop all tasks")

        if hasattr(self, "events"):
            self.events.stop()
        self.clock.loop.stop()
        self.clock.loop.run_forever()
        self.clock.loop.close()

    def _run_loop(self) -> None:    # pragma: no cover
        # Main machine run loop with when the default platform interface
        # specifies the MPF should control the main timer

        try:
            reason = self.clock.run(self.stop_future)
        except KeyboardInterrupt:
            print("Shutdown because of keyboard interrupts")
            return
        except BaseException as e:   # pylint: disable-msg=broad-except
            # this happens when receiving a signal
            self._crash_shutdown()
            self.log.exception("Loop exited with exception")
            report_crash(e, "run_loop", self.config)
            return

        if self._exception:
            self._crash_shutdown()

            print("Shutdown because of an exception:")
            try:
                raise self._exception['exception']
            except:     # noqa
                self.log.exception("Runtime Exception")
                report_crash(self._exception['exception'], "exception_during_run", self.config)
            self._exception = None
        else:
            self._do_stop()
            print("Shutdown reason: {}".format(reason))

    def _platform_stop(self) -> None:
        """Stop all platforms."""
        for hardware_platform in list(self.hardware_platforms.values()):
            hardware_platform.stop()

    def get_platform_sections(self, platform_section: str, overwrite: str) -> "SmartVirtualHardwarePlatform":
        """Return platform section."""
        assert self.default_platform is not None
        if overwrite == "drivers":
            return self.hardware_platforms[overwrite]

        if self.options['force_platform']:
            return self.default_platform

        if not overwrite:
            if self.config['hardware'][platform_section][0] != 'default':
                return self.hardware_platforms[self.config['hardware'][platform_section][0]]

            return self.default_platform

        try:
            return self.hardware_platforms[overwrite]
        except KeyError:
            raise AssertionError("Platform \"{}\" has not been loaded. Please add it to your \"hardware\" section.".
                                 format(overwrite))

    def register_boot_hold(self, hold: str) -> None:
        """Register a boot hold."""
        assert self.is_init_done is not None
        if self.is_init_done.is_set():
            raise AssertionError("Register hold after init_done")
        self._boot_holds.add(hold)

    def clear_boot_hold(self, hold: str) -> None:
        """Clear a boot hold."""
        assert self.is_init_done is not None
        if self.is_init_done.is_set():
            raise AssertionError("Clearing hold after init_done")
        self._boot_holds.remove(hold)
        self.debug_log('Clearing boot hold %s. Holds remaining: %s', hold, self._boot_holds)
        if not self._boot_holds:
            self.is_init_done.set()

    async def init_done(self) -> None:
        """Finish init.

        Called when init is done and all boot holds are cleared.
        """
        await self.events.post_async("init_done")
        '''event: init_done

        desc: Posted when the initial (one-time / boot) init phase is done. In
        other words, once this is posted, MPF is booted and ready to go.
        '''

        await self.reset()



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase


class TestSpinners(MpfTestCase):

    def get_config_file(self):
        return 'test_spinners.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/spinners/'

    def get_platform(self):
        return 'smart_virtual'

    def test_spinner_active_inactive(self):
        self.mock_event("spinner_spin1_active")
        self.mock_event("spinner_spin1_hit")
        self.mock_event("spinner_spin1_inactive")
        self.mock_event("spinner_spin1_idle")

        self.hit_and_release_switch("switch1")
        self.assertEventCalled("spinner_spin1_active")
        self.assertEventCalled("spinner_spin1_hit")

        self.advance_time_and_run(0.5)
        self.assertEventNotCalled("spinner_spin1_inactive")
        self.assertEventNotCalled("spinner_spin1_idle")

        self.advance_time_and_run(0.5)
        self.assertEventCalled("spinner_spin1_inactive")
        self.assertEventNotCalled("spinner_spin1_idle")

    def test_spinner_active_inactive_idle(self):
        self.mock_event("spinner_spin2_active")
        self.mock_event("spinner_spin2_hit")
        self.mock_event("spinner_spin2_inactive")
        self.mock_event("spinner_spin2_idle")

        self.hit_and_release_switch("switch2")
        self.assertEventCalled("spinner_spin2_active")
        self.assertEventCalled("spinner_spin2_hit")

        self.advance_time_and_run(0.1)
        self.assertEventNotCalled("spinner_spin2_inactive")
        self.assertEventNotCalled("spinner_spin2_idle")

        self.advance_time_and_run(0.4)
        self.assertEventCalled("spinner_spin2_inactive")
        self.assertEventNotCalled("spinner_spin2_idle")

        self.advance_time_and_run(0.5)
        self.assertEventNotCalled("spinner_spin2_idle")

        self.advance_time_and_run(0.4)
        self.assertEventCalled("spinner_spin2_idle")

    def test_single_spinner_hits(self):
        self.mock_event("spinner_spin1_active")
        self.mock_event("spinner_spin1_hit")
        self.mock_event("spinner_spin1_inactive")
        self.mock_event("spinner_spin1_idle")

        self.hit_and_release_switch("switch1")
        self.assertEventCalled("spinner_spin1_active")
        self.assertEventCalled("spinner_spin1_hit")
        self.assertEqual({"hits": 1, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin1_hit'])

        self.hit_and_release_switch("switch1")
        self.advance_time_and_run(0.5)
        self.assertEqual({"hits": 2, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin1_hit'])
        self.assertEventNotCalled("spinner_spin1_inactive")

        self.hit_and_release_switch("switch1")
        self.advance_time_and_run(0.5)
        self.assertEqual({"hits": 3, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin1_hit'])
        self.assertEventNotCalled("spinner_spin1_inactive")

        self.hit_and_release_switch("switch1")
        self.advance_time_and_run(0.5)
        self.assertEqual({"hits": 4, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin1_hit'])
        self.assertEventNotCalled("spinner_spin1_inactive")

        self.advance_time_and_run(0.5)
        self.assertEqual(1, self._events["spinner_spin1_active"])
        self.assertEqual(4, self._events["spinner_spin1_hit"])
        self.assertEventCalled("spinner_spin1_inactive")
        self.assertEventNotCalled("spinner_spin1_idle")

    def test_double_spinner_hits(self):
        self.mock_event("spinner_spin2_active")
        self.mock_event("spinner_spin2_hit")
        self.mock_event("spinner_spin2_foo_hit")
        self.mock_event("spinner_spin2_bar_hit")
        self.mock_event("spinner_spin2_inactive")
        self.mock_event("spinner_spin2_idle")

        self.hit_and_release_switch("switch2")
        self.assertEventCalled("spinner_spin2_active")
        self.assertEventCalled("spinner_spin2_hit")
        self.assertEventCalled("spinner_spin2_foo_hit")
        self.assertEqual({"hits": 1, "change": 1, "label": "foo"}, self._last_event_kwargs['spinner_spin2_hit'])
        self.assertEqual(1, self._events["spinner_spin2_foo_hit"])
        self.assertEqual(0, self._events["spinner_spin2_bar_hit"])

        self.hit_and_release_switch("switch3")
        self.advance_time_and_run(0.3)
        self.assertEqual({"hits": 2, "change": 1, "label": "bar"}, self._last_event_kwargs['spinner_spin2_hit'])
        self.assertEqual(1, self._events["spinner_spin2_foo_hit"])
        self.assertEqual(1, self._events["spinner_spin2_bar_hit"])

        self.hit_and_release_switch("switch3")
        self.advance_time_and_run(0.3)
        self.assertEqual({"hits": 3, "change": 1, "label": "bar"}, self._last_event_kwargs['spinner_spin2_hit'])
        self.assertEqual(1, self._events["spinner_spin2_foo_hit"])
        self.assertEqual(2, self._events["spinner_spin2_bar_hit"])

        self.hit_and_release_switch("switch2")
        self.advance_time_and_run(0.3)
        self.assertEqual({"hits": 4, "change": 1, "label": "foo"}, self._last_event_kwargs['spinner_spin2_hit'])
        self.assertEqual(2, self._events["spinner_spin2_foo_hit"])
        self.assertEqual(2, self._events["spinner_spin2_bar_hit"])
        self.assertEventNotCalled("spinner_spin2_inactive")

        self.advance_time_and_run(0.3)
        self.assertEventCalled("spinner_spin2_inactive")
        self.assertEventNotCalled("spinner_spin2_idle")

        # Re-activate the spinner before it goes idle
        self.hit_and_release_switch("switch2")
        self.advance_time_and_run(0.3)
        self.assertEqual({"hits": 1, "change": 1, "label": "foo"}, self._last_event_kwargs['spinner_spin2_hit'])
        self.assertEqual(3, self._events["spinner_spin2_foo_hit"])
        self.assertEqual(2, self._events["spinner_spin2_bar_hit"])
        # Active should be called again
        self.assertEqual(2, self._events["spinner_spin2_active"])
        self.assertEqual(1, self._events["spinner_spin2_inactive"])
        self.assertEventNotCalled("spinner_spin2_idle")

        self.advance_time_and_run(0.3)
        self.assertEqual(2, self._events["spinner_spin2_active"])
        self.assertEqual(2, self._events["spinner_spin2_inactive"])
        self.assertEventNotCalled("spinner_spin2_idle")

        self.advance_time_and_run(0.5)
        self.assertEventNotCalled("spinner_spin2_idle")

        self.advance_time_and_run(0.3)
        self.assertEventCalled("spinner_spin2_idle")

    def test_event_buffer(self):
        self.mock_event("spinner_spin_with_buffer_active")
        self.mock_event("spinner_spin_with_buffer_hit")

        self.hit_and_release_switch("switch5")
        self.assertEventCalled("spinner_spin_with_buffer_active")
        self.assertEventCalled("spinner_spin_with_buffer_hit")
        self.assertEqual({"hits": 1, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin_with_buffer_hit'])


        self.mock_event("spinner_spin_with_buffer_hit")
        self.hit_and_release_switch("switch5")
        self.hit_and_release_switch("switch5")
        self.hit_and_release_switch("switch5")
        self.assertEventNotCalled("spinner_spin_with_buffer_hit")

        self.advance_time_and_run(0.6)
        self.assertEventCalled("spinner_spin_with_buffer_active")
        self.assertEqual({"hits": 4, "change": 3, "label": None}, self._last_event_kwargs['spinner_spin_with_buffer_hit'])

        self.mock_event("spinner_spin_with_buffer_hit")
        self.hit_and_release_switch("switch5")
        self.hit_and_release_switch("switch5")
        self.assertEventNotCalled("spinner_spin_with_buffer_hit")

        self.advance_time_and_run(0.5)
        self.assertEventCalled("spinner_spin_with_buffer_active")
        self.assertEqual({"hits": 6, "change": 2, "label": None}, self._last_event_kwargs['spinner_spin_with_buffer_hit'])

    def test_reset_when_inactive_false(self):
        self.mock_event("spinner_spin3_active")
        self.mock_event("spinner_spin3_hit")
        self.mock_event("spinner_spin3_inactive")
        self.mock_event("spinner_spin3_idle")

        self.hit_and_release_switch("switch4")
        self.assertEventCalled("spinner_spin3_active")
        self.assertEventCalled("spinner_spin3_hit")
        self.assertEqual({"hits": 1, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin3_hit'])
        self.assertEventNotCalled("spinner_spin3_inactive")

        self.hit_and_release_switch("switch4")
        self.advance_time_and_run(0.3)
        self.assertEqual({"hits": 2, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin3_hit'])
        self.assertEventNotCalled("spinner_spin3_inactive")

        self.hit_and_release_switch("switch4")
        self.advance_time_and_run(0.3)
        self.assertEqual({"hits": 3, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin3_hit'])
        self.assertEventNotCalled("spinner_spin3_inactive")

        self.advance_time_and_run(0.3)
        self.assertEventCalled("spinner_spin3_inactive")
        self.assertEventNotCalled("spinner_spin3_idle")

        # Re-activate the spinner before it goes idle
        self.hit_and_release_switch("switch4")
        self.advance_time_and_run(0.3)
        self.assertEqual({"hits": 4, "change": 1, "label": None}, self._last_event_kwargs['spinner_spin3_hit'])
        # Active should be called again
        self.assertEqual(2, self._events["spinner_spin3_active"])
        self.assertEqual(1, self._events["spinner_spin3_inactive"])
        self.assertEventNotCalled("spinner_spin3_idle")

        self.advance_time_and_run(0.3)
        self.assertEqual(2, self._events["spinner_spin3_active"])
        self.assertEqual(2, self._events["spinner_spin3_inactive"])
        self.assertEventNotCalled("spinner_spin3_idle")

        self.advance_time_and_run(0.5)
        self.assertEventNotCalled("spinner_spin3_idle")

        self.advance_time_and_run(0.3)
        self.assertEventCalled("spinner_spin3_idle")
